from typing import Union

import abjad


class Score(abjad.Score):
    r"""Score."""

    def add_final_bar_line(
        self,
        bar_line: Union[str, abjad.BarLine] = "|.",
        *,
        to_each_voice: bool = False,
    ) -> None:
        r"""Adds a final bar line to all components of |abjad.Score|. Note that
            Auxjad adds this function as an extension method to |abjad.Score| (see
            usage below).

            Basic usage:
                >>> staff = abjad.Staff(r"c'4 d'4 e'4 f'4")
                >>> score = abjad.Score([staff])
                >>> abjad.show(score)

                ..  docs::

                    \new Score
                    <<
                        \new Staff
                        {
                            c'4
                            d'4
                            e'4
                            f'4
                        }
                    >>

                ..  figure:: ../_images/Score-fdO5TP6ff9.png

                >>> score.add_final_bar_line()
                >>> abjad.show(score)

                ..  docs::

                    \new Score
                    <<
                        \new Staff
                        {
                            c'4
                            d'4
                            e'4
                            f'4
                            \bar "|."
                        }
                    >>

                ..  figure:: ../_images/Score-SHok9MqNVh.png

            Multiple staves:
                Works with multiple staves:

                >>> staff_1 = abjad.Staff(r"c''1 d''1 e''1 f''1")
                >>> staff_2 = abjad.Staff(r"c'1 d'1 e'1 f'1")
                >>> score = abjad.Score([staff_1, staff_2])
                >>> score.add_final_bar_line()
                >>> abjad.show(score)

                ..  docs::

                    \new Score
                    <<
                        \new Staff
                        {
                            c''1
                            d''1
                            e''1
                            f''1
                            \bar "|."
                        }
                        \new Staff
                        {
                            c'1
                            d'1
                            e'1
                            f'1
                            \bar "|."
                        }
                    >>

                ..  figure:: ../_images/Score-oCR5QGUz6W.png

                Each stave will receive their own final bar line, which can be
                useful when part extracting:

                >>> abjad.show(staff_1)

                ..  docs::

                    \new Staff
                    {
                        c''1
                        d''1
                        e''1
                        f''1
                        \bar "|."
                    }

                ..  figure:: ../_images/Score-tTXh7GOAop.png

                >>> abjad.show(staff_2)

                ..  docs::

                    \new Staff
                    {
                        c'1
                        d'1
                        e'1
                        f'1
                        \bar "|."
                    }

                ..  figure:: ../_images/Score-MRWgOyiqt3.png

            ``to_each_voice``
                When multiple voices are present in a staff, the final bar line is
                added to the last voice only:

                >>> voice_1 = abjad.Voice(r"c''1 d''1 e''1 f''1")
                >>> voice_2 = abjad.Voice(r"c'2 d'2 e'2 f'2 g'2 a'2 b'2 c''2")
                >>> staff = abjad.Staff([voice_1, voice_2], simultaneous=True)
                >>> abjad.attach(abjad.LilyPondLiteral(r'\voiceOne'), voice_1)
                >>> abjad.attach(abjad.LilyPondLiteral(r'\voiceTwo'), voice_2)
                >>> score = auxjad.Score([staff])
                >>> score.add_final_bar_line()
                >>> abjad.show(score)

                ..  docs::

                    \new Score
                    <<
                        \new Staff
                        <<
                            \new Voice
                            {
                                \voiceOne
                                c''1
                                d''1
                                e''1
                                f''1
                            }
                            \new Voice
                            {
                                \voiceTwo
                                c'2
                                d'2
                                e'2
                                f'2
                                g'2
                                a'2
                                b'2
                                c''2
                                \bar "|."
                            }
                        >>
                    >>

                ..  figure:: ../_images/Score-8mnllVCEoL.png

                >>> abjad.show(voice_1)

                ..  docs::

                    \new Voice
                    {
                        \voiceOne
                        c''1
                        d''1
                        e''1
                        f''1
                    }

                ..  figure:: ../_images/Score-DPW54DDeyM.png

                >>> abjad.show(voice_2)

                ..  docs::

                    \new Voice
                    {
                        \voiceTwo
                        c'2
                        d'2
                        e'2
                        f'2
                        g'2
                        a'2
                        b'2
                        c''2
                        \bar "|."
                    }

                ..  figure:: ../_images/Score-ua8cZAITB5.png

                Setting ``to_each_voice`` to ``True`` will add a bar line to each
                voice in a staff:

                >>> voice_1 = abjad.Voice(r"c''1 d''1 e''1 f''1")
                >>> voice_2 = abjad.Voice(r"c'2 d'2 e'2 f'2 g'2 a'2 b'2 c''2")
                >>> staff = abjad.Staff([voice_1, voice_2], simultaneous=True)
                >>> abjad.attach(abjad.LilyPondLiteral(r'\voiceOne'), voice_1)
                >>> abjad.attach(abjad.LilyPondLiteral(r'\voiceTwo'), voice_2)
                >>> score = auxjad.Score([staff])
                >>> score.add_final_bar_line(to_each_voice=True)
                >>> abjad.show(voice_1)

                ..  docs::

                    \new Voice
                    {
                        \voiceOne
                        c''1
                        d''1
                        e''1
                        f''1
                        \bar "|."
                    }

                ..  figure:: ../_images/Score-adtLRR3v1W.png

                >>> abjad.show(voice_2)

                ..  docs::

                    \new Voice
                    {
                        \voiceTwo
                        c'2
                        d'2
                        e'2
                        f'2
                        g'2
                        a'2
                        b'2
                        c''2
                        \bar "|."
                    }

                ..  figure:: ../_images/Score-DqhspXgvIJ.png

            ..  warning::

                If voices do not end together then manually adding bar lines will
                be required:

                >>> voice_1 = abjad.Voice(r"c''1 d''1 e''1 f''1")
                >>> voice_2 = abjad.Voice(r"c'1 d'1 e'1")
                >>> staff = abjad.Staff([voice_1, voice_2], simultaneous=True)
                >>> abjad.attach(abjad.LilyPondLiteral(r'\voiceOne'), voice_1)
                >>> abjad.attach(abjad.LilyPondLiteral(r'\voiceTwo'), voice_2)
                >>> score = auxjad.Score([staff])
                >>> score.add_final_bar_line(to_each_voice=True)
                >>> abjad.show(score)

                ..  docs::

                    \new Score
                    <<
                        \new Staff
                        <<
                            \new Voice
                            {
                                \voiceOne
                                c''1
                                d''1
                                e''1
                                f''1
                                \bar "|."
                            }
                            \new Voice
                            {
                                \voiceTwo
                                c'1
                                d'1
                                e'1
                                \bar "|."
                            }
                        >>
                    >>

                ..  figure:: ../_images/Score-duPkS4pJGc.png

        argument:
            The default bar line is of type ``'|.'``. To change this behaviour,
            call this method with an argument of type :obj:`str` or |abjad.BarLine|
            with the desired bar line type:

            >>> staff_1 = abjad.Staff(r"c''1 d''1 e''1 f''1")
            >>> staff_2 = abjad.Staff(r"c'1 d'1 e'1 f'1")
            >>> score = auxjad.Score([staff_1, staff_2])
            >>> score.add_final_bar_line(abjad.BarLine(":|."))
            >>> abjad.show(score)

            ..  docs::

                \new Score
                <<
                    \new Staff
                    {
                        c''1
                        d''1
                        e''1
                        f''1
                        \bar ":|."
                    }
                    \new Staff
                    {
                        c'1
                        d'1
                        e'1
                        f'1
                        \bar ":|."
                    }
                >>

            ..  figure:: ../_images/Score-3JvhD9DKvE.png
        """
        if not isinstance(bar_line, (str, abjad.BarLine)):
            raise TypeError("argument must be 'str' or 'abjad.BarLine'")
        if not isinstance(to_each_voice, bool):
            raise TypeError("'to_each_voice' must be 'bool'")
        if isinstance(bar_line, str):
            bar_line = abjad.BarLine(bar_line)
        if not to_each_voice:
            for staff in abjad.Iteration(self).components(abjad.Staff):
                last_leaf = abjad._iterate._get_leaf(staff, -1)
                abjad.attach(bar_line, last_leaf, context="Staff")
        else:
            for staff in abjad.Iteration(self).components(abjad.Staff):
                voices = [voice for voice in abjad.Iteration(staff).components(abjad.Voice)]
                if len(voices) == 0:
                    last_leaf = abjad._iterate._get_leaf(staff, -1)
                    abjad.attach(bar_line, last_leaf, context="Staff")
                else:
                    for voice in abjad.Iteration(self).components(abjad.Voice):
                        last_leaf = abjad._iterate._get_leaf(voice, -1)
                        abjad.attach(bar_line, last_leaf, context="Voice")

    def add_double_bar_lines_before_time_signatures(
        self,
        *,
        to_each_voice: bool = False,
    ) -> None:
        r"""Adds double bar lines to all components of |abjad.Score| before
        every time signature changes. Note that Auxjad adds this function as an
        extension method to |abjad.Score| (see usage below).

        Basic usage:
            >>> staff = abjad.Staff(
            ...     r"\time 3/4 c'2. \time 4/4 d'1 e'1 \time 6/4 f'2. g'2."
            ... )
            >>> score = abjad.Score([staff])
            >>> abjad.show(score)

            ..  docs::

                \new Score
                <<
                    \new Staff
                    {
                        \time 3/4
                        c'2.
                        \time 4/4
                        d'1
                        e'1
                        \time 6/4
                        f'2.
                        g'2.
                    }
                >>

            ..  figure:: ../_images/Score-65l45tLipK.png

            >>> score.add_double_bar_lines_before_time_signatures()
            >>> abjad.show(score)

            ..  docs::

                \new Score
                <<
                    \new Staff
                    {
                        \time 3/4
                        c'2.
                        \bar "||"
                        \time 4/4
                        d'1
                        e'1
                        \bar "||"
                        \time 6/4
                        f'2.
                        g'2.
                    }
                >>

            ..  figure:: ../_images/Score-cjmOgHxSJH.png

        Multiple staves:
            Works with multiple staves:

            >>> staff_1 = abjad.Staff(
            ...     r"\time 3/4 c''2. \time 4/4 d''1 e''1 "
            ...     "\time 6/4 f''2. g''2."
            ... )
            >>> staff_2 = abjad.Staff(
            ...     r"\time 3/4 c'2. \time 4/4 d'1 e'1 \time 6/4 f'2. g'2."
            ... )
            >>> score = abjad.Score([staff_1, staff_2])
            >>> score.add_double_bar_lines_before_time_signatures()
            >>> abjad.show(score)

            ..  docs::

                \new Score
                <<
                    \new Staff
                    {
                        \time 3/4
                        c''2.
                        \bar "||"
                        \time 4/4
                        d''1
                        e''1
                        \bar "||"
                        \time 6/4
                        f''2.
                        g''2.
                    }
                    \new Staff
                    {
                        \time 3/4
                        c'2.
                        \bar "||"
                        \time 4/4
                        d'1
                        e'1
                        \bar "||"
                        \time 6/4
                        f'2.
                        g'2.
                    }
                >>

            ..  figure:: ../_images/Score-cnkUgFuYmI.png

            Each stave will receive their own double bar lines, which can be
            useful when part extracting:

            >>> abjad.show(staff_1)

            ..  docs::

                \new Staff
                {
                    \time 3/4
                    c''2.
                    \bar "||"
                    \time 4/4
                    d''1
                    e''1
                    \bar "||"
                    \time 6/4
                    f''2.
                    g''2.
                }

            ..  figure:: ../_images/Score-TZaC5C12yA.png

            >>> abjad.show(staff_2)

            ..  docs::

                \new Staff
                {
                    \time 3/4
                    c'2.
                    \bar "||"
                    \time 4/4
                    d'1
                    e'1
                    \bar "||"
                    \time 6/4
                    f'2.
                    g'2.
                }

            ..  figure:: ../_images/Score-wZgCbZi8JD.png

        ``to_each_voice``
            When multiple voices are present in a staff, double bar lines will
            be added to the last voice only:

            >>> voice_1 = abjad.Voice(
            ...     r"\time 3/4 c''2. \time 4/4 d''1 e''1 "
            ...     "\time 6/4 f''2. g''2."
            ... )
            >>> voice_2 = abjad.Voice(
            ...     r"\time 3/4 c'2. \time 4/4 d'1 e'1 \time 6/4 f'2. g'2."
            ... )
            >>> staff = abjad.Staff([voice_1, voice_2], simultaneous=True)
            >>> abjad.attach(abjad.LilyPondLiteral(r'\voiceOne'), voice_1)
            >>> abjad.attach(abjad.LilyPondLiteral(r'\voiceTwo'), voice_2)
            >>> score = auxjad.Score([staff])
            >>> score.add_double_bar_lines_before_time_signatures()
            >>> abjad.show(score)

            ..  docs::

                \new Score
                <<
                    \new Staff
                    <<
                        \new Voice
                        {
                            \voiceOne
                            \time 3/4
                            c''2.
                            \time 4/4
                            d''1
                            e''1
                            \time 6/4
                            f''2.
                            g''2.
                        }
                        \new Voice
                        {
                            \voiceTwo
                            \time 3/4
                            c'2.
                            \bar "||"
                            \time 4/4
                            d'1
                            e'1
                            \bar "||"
                            \time 6/4
                            f'2.
                            g'2.
                        }
                    >>
                >>

            ..  figure:: ../_images/Score-F6f5a6z3ly.png

            >>> abjad.show(voice_1)

            ..  docs::

                \new Voice
                {
                    \voiceOne
                    \time 3/4
                    c''2.
                    \time 4/4
                    d''1
                    e''1
                    \time 6/4
                    f''2.
                    g''2.
                }

            ..  figure:: ../_images/Score-FEAvA2mn1t.png

            >>> abjad.show(voice_2)

            ..  docs::

                \new Voice
                {
                    \voiceTwo
                    \time 3/4
                    c'2.
                    \bar "||"
                    \time 4/4
                    d'1
                    e'1
                    \bar "||"
                    \time 6/4
                    f'2.
                    g'2.
                }

            ..  figure:: ../_images/Score-QKhtZj82cG.png

            Setting ``to_each_voice`` to ``True`` will add a bar line to each
            voice in a staff:

            >>> voice_1 = abjad.Voice(
            ...     r"\time 3/4 c''2. \time 4/4 d''1 e''1 "
            ...     "\time 6/4 f''2. g''2."
            ... )
            >>> voice_2 = abjad.Voice(
            ...     r"\time 3/4 c'2. \time 4/4 d'1 e'1 \time 6/4 f'2. g'2."
            ... )
            >>> staff = abjad.Staff([voice_1, voice_2], simultaneous=True)
            >>> abjad.attach(abjad.LilyPondLiteral(r'\voiceOne'), voice_1)
            >>> abjad.attach(abjad.LilyPondLiteral(r'\voiceTwo'), voice_2)
            >>> score = auxjad.Score([staff])
            >>> score.add_double_bar_lines_before_time_signatures(
            ...     to_each_voice=True,
            ... )
            >>> abjad.show(voice_1)

            ..  docs::

                \new Voice
                {
                    \voiceOne
                    \time 3/4
                    c''2.
                    \bar "||"
                    \time 4/4
                    d''1
                    e''1
                    \bar "||"
                    \time 6/4
                    f''2.
                    g''2.
                }

            ..  figure:: ../_images/Score-n5x7JhOHy6.png

            >>> abjad.show(voice_2)

            ..  docs::

                \new Voice
                {
                    \voiceTwo
                    \time 3/4
                    c'2.
                    \bar "||"
                    \time 4/4
                    d'1
                    e'1
                    \bar "||"
                    \time 6/4
                    f'2.
                    g'2.
                }

            ..  figure:: ../_images/Score-tAXRWLGxMp.png

        Multi-measure rests:
            This method can handle multi-measure rests.

            >>> staff = abjad.Staff(
            ...     r"\time 3/4 R1 * 3/4 "
            ...     r"\time 4/4 R1 * 2 "
            ...     r"\time 6/4 R1 * 6/4 "
            ...     r"\time 4/4 R1"
            ... )
            >>> score = abjad.Score([staff])
            >>> score.add_double_bar_lines_before_time_signatures()
            >>> abjad.show(score)

            ..  docs::

                \new Score
                <<
                    \new Staff
                    {
                        \time 3/4
                        R1 * 3/4
                        \bar "||"
                        \time 4/4
                        R1 * 2
                        \bar "||"
                        \time 6/4
                        R1 * 3/2
                        \bar "||"
                        \time 4/4
                        R1
                    }
                >>

            ..  figure:: ../_images/Score-nwYMM8JAYR.png

        Input with bar lines:
            If the score already contains non-standard bar lines at points
            where a time signature changes, only those of types ``"|"`` and
            ``""`` will be replaced, keeping all others as they were.

            >>> staff = abjad.Staff(
            ...     r"R1 "
            ...     r"\time 3/4 c'2. "
            ...     r"\time 4/4 d'1 "
            ...     r"e'1 "
            ...     r"\time 6/4 f'2. g'2. "
            ...     r"\time 2/4 a'2"
            ... )
            >>> abjad.attach(abjad.BarLine('.|:'), staff[0])
            >>> abjad.attach(abjad.BarLine(':|.'), staff[1])
            >>> abjad.attach(abjad.BarLine('|'), staff[3])
            >>> abjad.attach(abjad.BarLine('!'), staff[5])
            >>> score = abjad.Score([staff])
            >>> score.add_double_bar_lines_before_time_signatures()
            >>> abjad.show(score)

            ..  docs::

                \new Score
                <<
                    \new Staff
                    {
                        R1
                        \bar ".|:"
                        \time 3/4
                        c'2.
                        \bar ":|."
                        \time 4/4
                        d'1
                        e'1
                        \bar "||"
                        \time 6/4
                        f'2.
                        g'2.
                        \bar "!"
                        \time 2/4
                        a'2
                    }
                >>

            ..  figure:: ../_images/Score-XVe9oSR8Gi.png
        """
        if not isinstance(to_each_voice, bool):
            raise TypeError("'to_each_voice' must be 'bool'")
        for staff in abjad.Iteration(self).components(abjad.Staff):
            voices = [voice for voice in abjad.Iteration(staff).components(abjad.Voice)]
            if len(voices) == 0:
                self._double_bar_line_adder(staff)
            elif not to_each_voice:
                self._double_bar_line_adder(voices[-1])
            else:
                for voice in voices:
                    self._double_bar_line_adder(voice)

    @staticmethod
    def _double_bar_line_adder(
        container: abjad.Container,
    ) -> None:
        r"""Goes through a container and adds double bar lines before each and
        every time signature change.
        """
        leaves = abjad.select(container).leaves()
        for i, leaf in enumerate(leaves[1:], 1):
            time_signature = abjad.get.indicator(leaf, abjad.TimeSignature)
            if time_signature is not None:
                bar_line = abjad.get.indicator(leaves[i - 1], abjad.BarLine)
                if bar_line is not None and bar_line.abbreviation in ("|", ""):
                    abjad.detach(abjad.BarLine, leaves[i - 1])
                    bar_line = None
                if bar_line is None:
                    abjad.attach(
                        abjad.BarLine("||"),
                        leaves[i - 1],
                        context="Voice",
                    )


### EXTENSION METHODS ###

abjad.Score.add_final_bar_line = Score.add_final_bar_line
abjad.Score.add_double_bar_lines_before_time_signatures = (
    Score.add_double_bar_lines_before_time_signatures
)



from functools import partial
from typing import Union

import abjad


def _group_consecutive_rests(
    logical_tie: abjad.LogicalTie,
    *,
    include_multimeasure_rests: bool = True,
) -> Union[abjad.Leaf, bool]:
    r"""Private function used by |auxjad.select.logical_selections()| in order
    to group consecutive ties together. If a logical tie is made out of a rest,
    this function returns the value ``True``, otherwise it returns the logical
    tie's head object itself. When this output is used with
    |abjad.select.group_by()|, consecutive ``True`` values will be grouped
    together, thus selecting consecutive rests as a single 'logical selection'.
    Meanwhile, non-rest logical ties will return their individual unique head
    object, maintaining their individual selections as is.
    """
    if isinstance(logical_tie.head, abjad.Rest):
        return True
    elif isinstance(logical_tie.head, abjad.MultimeasureRest) and include_multimeasure_rests:
        return True
    else:
        return logical_tie.head


def logical_selections(
    container: Union[abjad.Container, abjad.Selection],
    *,
    include_multimeasure_rests: bool = True,
) -> abjad.Selection:
    r"""Takes an |abjad.Container| (or child class).
    Returns the logical selections of a container, that is the logical
    ties but with consecutive rests grouped together. Return value is in the
    form of a |abjad.Selection| of |abjad.Selection|'s.

    Basic usage:
        Usage is similar to |abjad.select.logical_ties()|:

        >>> container = abjad.Container(r"c'4 ~ c'16 r8. r4.. d'16 ~ d'4")
        >>> logical_selections = auxjad.select.logical_selections(container)
        >>> for logical_selection in logical_selections:
        ...     print(logical_selection.leaves())
        Selection([Note("c'4"), Note("c'16")])
        Selection([Rest('r8.'), Rest('r4..')])
        Selection([Note("d'16"), Note("d'4")])

    ..  note::

        Auxjad automatically adds this function as an extension function to
        |abjad.select|. It can thus be used from either |auxjad.select|_ or
        |abjad.select| namespaces. Therefore, the two lines below are
        equivalent:

        >>> container = abjad.Container(r"c'4 ~ c'16 r8. r4.. d'16 ~ d'4")
        >>> for logical_sel in auxjad.select.logical_selections(container):
        ...     print(logical_sel.leaves())
        Selection([Note("c'4"), Note("c'16")])
        Selection([Rest('r8.'), Rest('r4..')])
        Selection([Note("d'16"), Note("d'4")])
        >>> for logical_sel in abjad.select.logical_selections(container):
        ...     print(logical_sel.leaves())
        Selection([Note("c'4"), Note("c'16")])
        Selection([Rest('r8.'), Rest('r4..')])
        Selection([Note("d'16"), Note("d'4")])

    ``include_multimeasure_rests``:
        By default, this function groups multi-measure rests with the previous
        pitched leaf as a single logical selection.

        >>> container = abjad.Container(r"c'2. ~ c'16 r8. R1 r4.. d'16 ~ d'2.")
        >>> logical_selections = auxjad.select.logical_selections(container)
        >>> for logical_selection in logical_selections:
        ...     print(logical_selection.leaves())
        Selection([Note("c'2."), Note("c'16")])
        Selection([Rest('r8.'), MultimeasureRest('R1'), Rest('r4..')])
        Selection([Note("d'16"), Note("d'2.")])

        To treat multi-measure rests as their own logical selections, set
        ``include_multimeasure_rests`` to ``False`` (default value is
        ``True``).

        >>> container = abjad.Container(r"c'2. ~ c'16 r8. R1 r4.. d'16 ~ d'2.")
        >>> logical_selections = auxjad.select.logical_selections(
        ...     container,
        ...     include_multimeasure_rests=False,
        ... )
        >>> for logical_selection in logical_selections:
        ...     print(logical_selection.leaves())
        Selection([Note("c'2."), Note("c'16")])
        Selection([Rest('r8.')])
        Selection([MultimeasureRest('R1')])
        Selection([Rest('r4..')])
        Selection([Note("d'16"), Note("d'2.")])
    """
    if not isinstance(container, (abjad.Container, abjad.Selection)):
        raise TypeError("Argument must be 'abjad.Container', 'abjad.Selection', or child classes")
    if isinstance(container, abjad.Container):
        if not abjad.select(container).leaves().are_contiguous_logical_voice():
            raise ValueError("Argument must be contiguous logical voice")

    logical_ties = abjad.select(container).logical_ties()
    return logical_ties.group_by(
        partial(
            _group_consecutive_rests,
            include_multimeasure_rests=include_multimeasure_rests,
        )
    )



import random
from typing import Optional, Union

import abjad

from .. import mutate
from .TenneySelector import TenneySelector


class PitchRandomiser:
    r"""Takes an input |abjad.Container| (or child class) together with a  list
    of pitches and randomises the container's pitch content using the list of
    pitches. The pitches can be of type :obj:`list`, :obj:`tuple`, :obj:`str`,
    or |abjad.PitchSegment|.

    Basic usage:
        Calling the object will output a selection of the input container with
        randomised pitches. Pitches are randomly selected from :attr:`pitches`.

        >>> container = abjad.Container(r"\time 4/4 c'4 d'4 e'4 f'4")
        >>> abjad.show(container)

        ..  docs::

            {
                c'4
                d'4
                e'4
                f'4
            }

        ..  figure:: ../_images/PitchRandomiser-PxSLuwtgn9.png

        >>> pitches = r"fs' gs' a' b' cs''"
        >>> randomiser = auxjad.PitchRandomiser(container,
        ...                                     pitches,
        ...                                     )
        >>> notes = randomiser()
        >>> staff = abjad.Staff(notes)
        >>> abjad.show(staff)

        ..  docs::

            \new Staff
            {
                \time 4/4
                a'4
                b'4
                cs''4
                fs'4
            }

        ..  figure:: ../_images/PitchRandomiser-134lqskbb6o.png

        >>> notes = randomiser()
        >>> staff = abjad.Staff(notes)
        >>> abjad.show(staff)

        ..  docs::

            \new Staff
            {
                \time 4/4
                b'4
                b'4
                b'4
                fs'4
            }

        ..  figure:: ../_images/PitchRandomiser-z66g1fy8nm8.png

        To get the result of the last operation, use the property
        :attr:`current_window`.

        >>> notes = randomiser.current_window
        >>> staff = abjad.Staff(notes)
        >>> abjad.show(staff)

        ..  docs::

            \new Staff
            {
                \time 4/4
                b'4
                b'4
                b'4
                fs'4
            }

        ..  figure:: ../_images/PitchRandomiser-x0e6yduogh.png

    ..  warning::

        Unlike the other classes in Auxjad, the very first call of this class
        will already process the initial container. To disable this behaviour
        and output the initial container once before randomising its pitches,
        initialise the class with the keyword argument
        :attr:`process_on_first_call` set to ``False``.

        >>> container = abjad.Container(r"c'4 d'4 e'4 f'4")
        >>> pitches = r"fs' gs' a' b'"
        >>> randomiser = auxjad.PitchRandomiser(container,
        ...                                     pitches,
        ...                                     process_on_first_call=False,
        ...                                     )
        >>> notes = randomiser()
        >>> staff = abjad.Staff(notes)
        >>> abjad.show(staff)

        ..  docs::

            \new Staff
            {
                c'4
                d'4
                e'4
                f'4
            }

        ..  figure:: ../_images/PitchRandomiser-640x6vsjwtk.png

        >>> notes = randomiser()
        >>> staff = abjad.Staff(notes)
        >>> abjad.show(staff)

        ..  docs::

            \new Staff
            {
                b'4
                fs'4
                gs'4
                fs'4
            }

        ..  figure:: ../_images/PitchRandomiser-vsjdj8gkanj.png

    :func:`len()`:
        Applying the :func:`len()` function to the randomiser will return the
        number of pitches in :attr:`pitches`.

        >>> container = abjad.Container(r"c'4 d'4 e'4 f'4")
        >>> pitches = r"fs' gs' a' b'"
        >>> randomiser = auxjad.PitchRandomiser(container,
        ...                                     pitches,
        ...                                     )
        >>> len(randomiser)
        4

        >>> container = abjad.Container(r"c'4 d'4 e'4 f'4")
        >>> pitches = [6, 7, 8, 9, 10, 11, 12]
        >>> randomiser = auxjad.PitchRandomiser(container,
        ...                                     pitches,
        ...                                     )
        >>> len(randomiser)
        7

    Arguments and properties:
        This class has many keyword arguments, all of which can be altered
        after instantiation using properties with the same names as shown
        below. :attr:`weights` takes a :obj:`list` of :obj:`int`'s or
        :obj:`float`'s representing the weight of each pitch from
        :attr:`pitches` (their lengths must also match).
        :attr:`omit_time_signatures` will remove all time signatures from the
        output (both are ``False`` by default). :attr:`process_on_first_call`
        to ``True`` and the random pitch process will be applied on the very
        first call. Setting :attr:`use_tenney_selector` to ``True`` will make
        the randomiser use :class:`auxjad.TenneySelector` for the random
        selection instead of :func:`random.choices()` (default is ``False``).

        >>> container = abjad.Container(r"c'4 d'4 e'4 f'4")
        >>> randomiser = auxjad.PitchRandomiser(
        ...     container,
        ...     pitches=r"a b cs' ds' e'",
        ...     weights=[1.0, 2.0, 1.0, 1.5, 1.3],
        ...     omit_time_signatures=True,
        ...     process_on_first_call=True,
        ...     use_tenney_selector=True,
        ... )
        >>> randomiser.pitches
        <a b cs' ds' e'>
        >>> randomiser.weights
        [1.0, 2.0, 1.0, 1.5, 1.3]
        >>> randomiser.omit_time_signatures
        True
        >>> randomiser.process_on_first_call
        True
        >>> randomiser.use_tenney_selector
        True

        Use the properties below to change these values after initialisation.

        >>> randomiser.pitches = abjad.PitchSegment(r"c' d' e' f'")
        >>> randomiser.weights = [1, 2, 5, 8]
        >>> randomiser.omit_time_signatures = False
        >>> randomiser.process_on_first_call = False
        >>> randomiser.use_tenney_selector = False
        >>> randomiser.pitches
        <c' d' e' f'>
        >>> randomiser.weights
        [1, 2, 5, 8]
        >>> randomiser.omit_time_signatures
        False
        >>> randomiser.process_on_first_call
        False
        >>> randomiser.use_tenney_selector
        False

    Rests:
        Only pitched logical ties are randomised, rests are left untouched.

        >>> container = abjad.Container(r"c'8. d'4 r8 r8. e'16 f'8.")
        >>> pitches = [6, 7, 8, 9, 10, 11]
        >>> randomiser = auxjad.PitchRandomiser(container,
        ...                                     pitches,
        ...                                     )
        >>> notes = randomiser()
        >>> staff = abjad.Staff(notes)
        >>> abjad.show(staff)

        ..  docs::

            \new Staff
            {
                bf'8.
                af'4
                r8
                r8.
                bf'16
                a'8.
            }

        ..  figure:: ../_images/PitchRandomiser-u4294ozm92.png

    Chords:
        Each note of a chord is randomised too.

        >>> container = abjad.Container(
        ...     r"<c' e' g'>8. d'4 r8 r8. e'16 <f' a'>8."
        ... )
        >>> pitches = [6, 7, 8, 9, 10, 11]
        >>> randomiser = auxjad.PitchRandomiser(container,
        ...                                     pitches,
        ...                                     )
        >>> notes = randomiser()
        >>> staff = abjad.Staff(notes)
        >>> abjad.show(staff)

        ..  docs::

            \new Staff
            {
                <fs' g' bf'>8.
                b'4
                r8
                r8.
                bf'16
                <fs' bf'>8.
            }

        ..  figure:: ../_images/PitchRandomiser-318eldj7tzc.png

        The number of notes in a chord stay the same unless there are fewer
        pitches available in :attr:`pitches`.

        >>> container = abjad.Container(
        ...     r"<c' e' g' a'>2 <cs' ds' e' f' g' a' b'>2"
        ... )
        >>> pitches = [6, 7, 8]
        >>> randomiser = auxjad.PitchRandomiser(container,
        ...                                     pitches,
        ...                                     )
        >>> notes = randomiser()
        >>> staff = abjad.Staff(notes)
        >>> abjad.show(staff)

        ..  docs::

            \new Staff
            {
                <fs' g' af'>2
                <fs' g' af'>2
            }

        ..  figure:: ../_images/PitchRandomiser-cciyee49qrj.png

    :attr:`use_tenney_selector`:
        Setting :attr:`use_tenney_selector` to ``True`` will make the
        randomiser use :class:`auxjad.TenneySelector` for the random selection
        instead of :func:`random.choices()` (default is ``False``).
        :class:`auxjad.TenneySelector` will raise the chance of a pitch being
        selected the longer it hasn't been selected, and will forbid immediate
        repetitions of pitches. See its documentation for more information.

        >>> container = abjad.Container(r"c'8 d'8 e'8 f'8 g'8 a'8 b'8 c'8")
        >>> pitches = r"fs' gs' a' b'"
        >>> randomiser = auxjad.PitchRandomiser(container,
        ...                                     pitches,
        ...                                     use_tenney_selector=True,
        ...                                     )
        >>> notes = randomiser()
        >>> staff = abjad.Staff(notes)
        >>> abjad.show(staff)

        ..  docs::

            \new Staff
            {
                fs'8
                a'8
                fs'8
                gs'8
                a'8
                b'8
                gs'8
                fs'8
            }

        ..  figure:: ../_images/PitchRandomiser-tmxllu13pa8.png

    :attr:`weights`:
        Individual pitches can have different weights, defined by the
        :attr:`weights` property. It takes a :obj:`list` of :obj:`float`'s or
        :obj:`int`'s.

        >>> container = abjad.Container(r"c'8 d'8 e'8 f'8 g'8 a'8 b'8 c'8")
        >>> pitches = r"fs' gs' a' b'"
        >>> randomiser = auxjad.PitchRandomiser(container,
        ...                                     pitches,
        ...                                     weights=[5.0, 2.0, 1.5, 1.0],
        ...                                     )
        >>> notes = randomiser()
        >>> staff = abjad.Staff(notes)
        >>> abjad.show(staff)

        ..  docs::

            \new Staff
            {
                b'8
                fs'8
                gs'8
                gs'8
                gs'8
                fs'8
                fs'8
                a'8
            }

        ..  figure:: ../_images/PitchRandomiser-zkvbzd1brgq.png

    :attr:`weights` and :attr:`use_tenney_selector`:
        Non-uniform :attr:`weights` can also be used when
        :attr:`use_tenney_selector` is set to ``True``.

        >>> container = abjad.Container(r"c'8 d'8 e'8 f'8 g'8 a'8 b'8 c'8")
        >>> pitches = r"fs' gs' a' b'"
        >>> randomiser = auxjad.PitchRandomiser(container,
        ...                                     pitches,
        ...                                     weights=[5.0, 2.0, 1.5, 1.0],
        ...                                     use_tenney_selector=True,
        ...                                     )
        >>> notes = randomiser()
        >>> staff = abjad.Staff(notes)
        >>> abjad.show(staff)

        ..  docs::

            \new Staff
            {
                fs'8
                gs'8
                a'8
                fs'8
                gs'8
                a'8
                gs'8
                b'8
            }

        ..  figure:: ../_images/PitchRandomiser-cq1nobkjozg.png

    Resetting :attr:`weights`:
        Setting :attr:`weights` to ``None`` will reset it back to a uniform
        distribution.

        >>> container = abjad.Container(r"c'8 d'8 e'8 f'8 g'8 a'8 b'8 c'8")
        >>> pitches = r"fs' gs' a' b'"
        >>> randomiser = auxjad.PitchRandomiser(container,
        ...                                     pitches,
        ...                                     weights=[100.0, 1.0, 1.0, 1.0],
        ...                                     )
        >>> notes = randomiser()
        >>> staff = abjad.Staff(notes)
        >>> abjad.show(staff)

        ..  docs::

            \new Staff
            {
                fs'8
                fs'8
                fs'8
                fs'8
                fs'8
                fs'8
                fs'8
                fs'8
            }

        ..  figure:: ../_images/PitchRandomiser-wtl5o15q5qp.png

        >>> randomiser.weights = None
        >>> notes = randomiser()
        >>> staff = abjad.Staff(notes)
        >>> abjad.show(staff)

        ..  docs::

            \new Staff
            {
                fs'8
                fs'8
                b'8
                gs'8
                gs'8
                a'8
                fs'8
                b'8
            }

        ..  figure:: ../_images/PitchRandomiser-4bqe6cbawg6.png

    Changing :attr:`pitches`:
        When using a custom :obj:`list` of :attr:`weights`, changing the
        :attr:`pitches` to a series of new values with the same length will
        preserve the :attr:`weights` values. If on the other hand
        :attr:`pitches` changes in length, :attr:`weights` is reset to ``None``
        (i.e. uniform distribution).

        >>> container = abjad.Container(r"c'8 d'8 e'8 f'8 g'8 a'8 b'8 c'8")
        >>> pitches = r"fs' gs' a' b'"
        >>> randomiser = auxjad.PitchRandomiser(container,
        ...                                     pitches,
        ...                                     weights=[100.0, 1.0, 1.0, 1.0],
        ...                                     )
        >>> randomiser.pitches = r"c'' d'' e'' f''"
        >>> randomiser.pitches
        <c'' d'' e'' f''>
        >>> randomiser.weights
        [100.0, 1.0, 1.0, 1.0]
        >>> randomiser.pitches = r"c'' d'' e'' f'' g'' a'' b''"
        >>> randomiser.pitches
        <c'' d'' e'' f'' g'' a'' b''>
        >>> randomiser.weights
        None

    ..  error::

        Note that :attr:`weights` must always have the same length as
        :attr:`pitches`, otherwise a :exc:`ValueError` exception will be
        raised.

        >>> container = abjad.Container(r"c'4 d'4 e'4 f'4")
        >>> pitches = r"fs' gs' a' b'"
        >>> weights = [1, 1, 5, 2, 3, 4, 8]
        >>> auxjad.PitchRandomiser(container, pitches, weights=weights)
        ValueError: 'weights' must have the same length as 'pitches'

    :meth:`output_n`:
        To output several randomised containers at once, use the method
        :meth:`output_n`, inputting the desired number of iterations.

        >>> container = abjad.Container(r"c'4 ~ c'16 r8. d'4 e'8. r16")
        >>> pitches = [6, 7, 8, 9, 10]
        >>> randomiser = auxjad.PitchRandomiser(container,
        ...                                     pitches,
        ...                                     )
        >>> notes = randomiser.output_n(3)
        >>> staff = abjad.Staff(notes)
        >>> abjad.show(staff)

        ..  docs::

            \new Staff
            {
                a'4
                ~
                a'16
                r8.
                g'4
                fs'8.
                r16
                g'4
                ~
                g'16
                r8.
                g'4
                fs'8.
                r16
                bf'4
                ~
                bf'16
                r8.
                a'4
                af'8.
                r16
            }

        ..  figure:: ../_images/PitchRandomiser-fvwaaz3vgi.png

    Indicators:
        This class preserves indicators.

        >>> container = abjad.Container(
        ...     r"c'4\p\< ~ c'8. d'16-.\f e'4--\pp f'8.( g'16)"
        ... )
        >>> pitches = [6, 7, 8, 9, 10, 11, 12]
        >>> randomiser = auxjad.PitchRandomiser(container,
        ...                                     pitches,
        ...                                     )
        >>> notes = randomiser()
        >>> staff = abjad.Staff(notes)
        >>> abjad.show(staff)

        ..  docs::

            \new Staff
            {
                a'4
                \p
                \<
                ~
                a'8.
                c''16
                \f
                - \staccato
                af'4
                \pp
                - \tenuto
                a'8.
                (
                bf'16
                )
            }

        ..  figure:: ../_images/PitchRandomiser-2e647eng8yc.png

    Example:
        This class also preserves the time signature structure.

        >>> container = abjad.Container(
        ...     r"\time 3/4 c'4 d'2 \time 2/4 e'8 f'8 g'8 a'8"
        ... )
        >>> pitches = r"fs' gs' a' b'"
        >>> randomiser = auxjad.PitchRandomiser(container,
        ...                                     pitches,
        ...                                     )
        >>> notes = randomiser()
        >>> staff = abjad.Staff(notes)
        >>> abjad.show(staff)

        ..  docs::

            \new Staff
            {
                \time 3/4
                a'4
                fs'2
                \time 2/4
                gs'8
                gs'8
                a'8
                gs'8
            }

        ..  figure:: ../_images/PitchRandomiser-c9t7r3thrqg.png

    :attr:`omit_time_signatures`:
        To omit time signatures altogether, set :attr:`omit_time_signatures` to
        ``True`` (default is ``False``).

        >>> container = abjad.Container(
        ...     r"\time 3/4 c'4 d'2 \time 2/4 e'8 f'8 g'8 a'8"
        ... )
        >>> pitches = r"fs' gs' a' b'"
        >>> randomiser = auxjad.PitchRandomiser(container,
        ...                                     pitches,
        ...                                     omit_time_signatures=True,
        ...                                     )
        >>> notes = randomiser()
        >>> staff = abjad.Staff(notes)
        >>> abjad.show(staff)

        ..  docs::

            \new Staff
            {
                a'4
                fs'2
                gs'8
                gs'8
                a'8
                gs'8
            }

        ..  figure:: ../_images/PitchRandomiser-mwruvbmgu3o.png

    Using as iterator:
        The instances of this class can also be used as an iterator, which can
        then be used in a for loop. Note that unlike the method
        :meth:`output_n`, time signatures are added to each window returned by
        the randomiser. Use the function
        |auxjad.mutate.remove_repeated_time_signatures()| to clean the output
        when using this class in this way. It is also important to note that a
        ``break`` statement is needed when using this class as an iterator. The
        reason is that pitch randomisation is a process that can happen
        indefinitely (unlike some of the other classes in this library).

        >>> container = abjad.Container(r"\time 3/4 c'4 d'4 e'4")
        >>> pitches = r"fs' gs' a' b' cs''"
        >>> randomiser = auxjad.PitchRandomiser(container,
        ...                                     pitches,
        ...                                     )
        >>> staff = abjad.Staff()
        >>> for window in randomiser:
        ...     staff.append(window)
        ...     if abjad.get.duration(staff) == abjad.Duration((9, 4)):
        ...         break
        >>> auxjad.mutate.remove_repeated_time_signatures(staff)
        >>> abjad.show(staff)

        ..  docs::

            \new Staff
            {
                \time 3/4
                a'4
                b'4
                gs'4
                fs'4
                a'4
                b'4
                a'4
                gs'4
                cs''4
            }

        ..  figure:: ../_images/PitchRandomiser-2c9zzip8tpc.png

    ..  tip::

        The functions |auxjad.mutate.remove_repeated_dynamics()| and
        |auxjad.mutate.reposition_clefs()| can be used to clean the output
        and remove repeated dynamics and unnecessary clef changes.
    """

    ### CLASS VARIABLES ###

    __slots__ = (
        "_contents",
        "_pitches",
        "_weights",
        "_omit_time_signatures",
        "_process_on_first_call",
        "_use_tenney_selector",
        "_tenney_selector",
        "_current_window",
        "_is_first_window",
    )

    ### INITIALISER ###

    def __init__(
        self,
        contents: abjad.Container,
        pitches: Union[
            list[Union[int, float, str, abjad.Pitch]],
            tuple[Union[int, float, str, abjad.Pitch]],
            str,
            abjad.PitchSegment,
        ],
        *,
        weights: Optional[list] = None,
        omit_time_signatures: bool = False,
        process_on_first_call: bool = True,
        use_tenney_selector: bool = False,
    ) -> None:
        self.contents = contents
        self._weights = []
        self.pitches = pitches
        self.weights = weights
        self.omit_time_signatures = omit_time_signatures
        self.process_on_first_call = process_on_first_call
        self.use_tenney_selector = use_tenney_selector
        self._is_first_window = True

    ### SPECIAL METHODS ###

    def __repr__(self) -> str:
        r"""Returns interpreter representation of :attr:`pitches`."""
        return repr(self._pitches)

    def __len__(self) -> int:
        r"""Returns the number of available :attr:`pitches`."""
        return len(self._pitches)

    def __call__(self) -> abjad.Selection:
        r"""Calls the randomise process, returning an |abjad.Selection|"""
        self._randomise()
        return self.current_window

    def __next__(self) -> abjad.Selection:
        r"""Calls the randomise process for one iteration, returning an
        |abjad.Selection|.
        """
        return self.__call__()

    def __iter__(self) -> None:
        r"""Returns an iterator, allowing instances to be used as iterators."""
        return self

    ### PUBLIC METHODS ###

    def output_n(
        self,
        n: int,
    ) -> abjad.Selection:
        r"""Goes through ``n`` iterations of the pitch randomisation process
        and outputs a single |abjad.Selection|.
        """
        if not isinstance(n, int):
            raise TypeError("argument must be 'int'")
        if n < 1:
            raise ValueError("argument must be greater than zero")
        dummy_container = abjad.Container()
        for _ in range(n):
            dummy_container.append(self.__call__())
        mutate.remove_repeated_time_signatures(dummy_container[:])
        mutate.remove_repeated_dynamics(dummy_container[:])
        output = dummy_container[:]
        dummy_container[:] = []
        return output

    ### PRIVATE METHODS ###

    def _randomise(self) -> abjad.Selection:
        r"""Randomises pitches of :attr:`contents`."""
        if self._is_first_window and not self._process_on_first_call:
            self._is_first_window = False
        else:
            self._rewrite_pitches()

    def _rewrite_pitches(self) -> None:
        r"""Rewrites the pitches of the current window."""
        dummy_container = abjad.mutate.copy(self._contents)
        logical_ties = abjad.select(dummy_container).logical_ties()
        for logical_tie in logical_ties:
            if isinstance(logical_tie[0], abjad.Note):
                pitch = self._pick_random_pitch()
                for leaf in logical_tie:
                    leaf.written_pitch = pitch
            elif isinstance(logical_tie[0], abjad.Chord):
                chord_n = len(logical_tie[0].written_pitches)
                if chord_n > self.__len__():
                    pitches = self._pitches
                else:
                    pitches = []
                    while len(pitches) < chord_n:
                        pitch = self._pick_random_pitch()
                        if pitch not in pitches:
                            pitches.append(pitch)
                for leaf in logical_tie:
                    leaf.written_pitches = pitches
        # output
        self._is_first_window = False
        self._current_window = dummy_container[:]
        dummy_container[:] = []

    def _pick_random_pitch(self) -> abjad.Pitch:
        r"""Random pitch selector, using either :func:`random.choices()` or
        :class:`auxjad.TenneySelector`.
        """
        if not self._use_tenney_selector:
            return random.choices(
                self._pitches,
                weights=self._weights,
            )[0]
        else:
            return self._tenney_selector()

    @staticmethod
    def _remove_all_time_signatures(
        container: abjad.Container,
    ) -> None:
        r"""Removes all time signatures of an |abjad.Container|."""
        for leaf in abjad.select(container).leaves():
            if abjad.get.effective(leaf, abjad.TimeSignature):
                abjad.detach(abjad.TimeSignature, leaf)

    ### PUBLIC PROPERTIES ###

    @property
    def contents(self) -> abjad.Container:
        r"""The |abjad.Container| to be shuffled."""
        return abjad.mutate.copy(self._contents)

    @contents.setter
    def contents(
        self,
        contents: abjad.Container,
    ) -> None:
        if not isinstance(contents, abjad.Container):
            raise TypeError("'contents' must be 'abjad.Container' or child class")
        if not abjad.select(contents).leaves().are_contiguous_logical_voice():
            raise ValueError("'contents' must be contiguous logical voice")
        if isinstance(contents, abjad.Score):
            self._contents = abjad.mutate.copy(contents[0])
        elif isinstance(contents, abjad.Tuplet):
            self._contents = abjad.Container([abjad.mutate.copy(contents)])
        else:
            self._contents = abjad.mutate.copy(contents)
        dummy_container = abjad.mutate.copy(contents)
        self._current_window = dummy_container[:]
        dummy_container[:] = []
        self._is_first_window = True

    @property
    def pitches(self) -> abjad.PitchSegment:
        r"""Pitches available for the randomiser."""
        return self._pitches

    @pitches.setter
    def pitches(
        self,
        pitches: Union[
            list[Union[int, float, str, abjad.Pitch]],
            tuple[Union[int, float, str, abjad.Pitch]],
            str,
            abjad.PitchSegment,
        ],
    ) -> None:
        if not isinstance(pitches, (list, tuple, str, abjad.PitchSegment)):
            raise TypeError("'pitches' must be 'list', 'tuple', 'str', or 'abjad.PitchSegment'")
        if isinstance(pitches, (list, tuple, str)):
            self._pitches = abjad.PitchSegment(pitches)
        else:
            self._pitches = pitches
        pitch_list = [pitch for pitch in self._pitches]
        self._tenney_selector = TenneySelector(pitch_list)
        if self._weights is not None:
            if len(pitch_list) != len(self._weights):
                self.weights = None

    @property
    def weights(self) -> list[Union[float, int]]:
        r"""The :obj:`list` with weights for each element of :attr:`pitches`"""
        return self._weights

    @weights.setter
    def weights(
        self,
        weights: Optional[list[Union[float, int]]],
    ) -> None:
        if weights is not None:
            if not isinstance(weights, list):
                raise TypeError("'weights' must be 'list'")
            if not self.__len__() == len(weights):
                raise ValueError("'weights' must have the same length as 'pitches'")
            if not all(isinstance(weight, (int, float)) for weight in weights):
                raise TypeError("'weights' elements must be 'int' or 'float'")
            self._weights = weights[:]
        else:
            self._weights = None
        self._tenney_selector.weights = self._weights

    @property
    def omit_time_signatures(self) -> bool:
        r"""When ``True``, the output will contain no time signatures."""
        return self._omit_time_signatures

    @omit_time_signatures.setter
    def omit_time_signatures(
        self,
        omit_time_signatures: bool,
    ) -> None:
        if not isinstance(omit_time_signatures, bool):
            raise TypeError("'omit_time_signatures' must be 'bool'")
        self._omit_time_signatures = omit_time_signatures

    @property
    def process_on_first_call(self) -> bool:
        r"""If ``True`` then :attr:`contents` will be processed in the very
        first call.
        """
        return self._process_on_first_call

    @process_on_first_call.setter
    def process_on_first_call(
        self,
        process_on_first_call: bool,
    ) -> None:
        if not isinstance(process_on_first_call, bool):
            raise TypeError("'process_on_first_call' must be 'bool'")
        self._process_on_first_call = process_on_first_call

    @property
    def use_tenney_selector(self) -> bool:
        r"""If ``True`` then the pitches will be selected using
        :class:`auxjad.TenneySelector`, otherwise they are chosen using a
        uniform random distribution.
        """
        return self._use_tenney_selector

    @use_tenney_selector.setter
    def use_tenney_selector(
        self,
        use_tenney_selector: bool,
    ) -> None:
        if not isinstance(use_tenney_selector, bool):
            raise TypeError("'use_tenney_selector' must be 'bool'")
        self._use_tenney_selector = use_tenney_selector

    @property
    def current_window(self) -> abjad.Selection:
        r"""Read-only property, returns the result of the last operation."""
        current_window = abjad.mutate.copy(self._current_window)
        if self._omit_time_signatures:
            self._remove_all_time_signatures(current_window)
        return current_window



import random
from typing import Optional, Union

import abjad

from .. import get, mutate


class Echoer:
    r"""Takes an |abjad.Container| (or child class) as input and, using it as
    reference, gradually lowers all dynamics, removing notes that are below a
    given threshold, returning the output as an |abjad.Selection|.

    Basic usage:
        Calling the object will return an |abjad.Selection| generated by the
        echoing process. Each call of the object will apply the echoing process
        to the previous result. By default, the container will be faded out
        (that is, its notes will be gradually removed one by one). Note that,
        by default, the first call in fade out mode outputs the initial
        container, with subsequent calls replacing leaves for rests.

        >>> container = abjad.Container(r"c'4\mf d'4\mp e'\p f'\pp")
        >>> abjad.show(container)

        ..  docs::

            {
                c'4
                \mf
                d'4
                \mp
                e'4
                \p
                f'4
                \pp
            }

        ..  figure:: ../_images/Echoer-jVGtGRdQev.png

        >>> echoer = auxjad.Echoer(container)
        >>> notes = echoer()
        >>> staff = abjad.Staff(notes)
        >>> abjad.show(staff)

        ..  docs::

            \new Staff
            {
                \time 4/4
                c'4
                \mf
                d'4
                \mp
                e'4
                \p
                f'4
                \pp
            }

        ..  figure:: ../_images/Echoer-4e4pnLJzwI.png

        >>> notes = echoer()
        >>> staff = abjad.Staff(notes)
        >>> abjad.show(staff)

        ..  docs::

            \new Staff
            {
                \time 4/4
                c'4
                \mp
                d'4
                \p
                e'4
                \pp
                f'4
                \ppp
            }

        ..  figure:: ../_images/Echoer-4bebflmnff9.png

        >>> notes = echoer()
        >>> staff = abjad.Staff(notes)
        >>> abjad.show(staff)

        ..  docs::

            \new Staff
            {
                \time 4/4
                c'4
                \p
                d'4
                \pp
                e'4
                \ppp
                r4
            }

        ..  figure:: ../_images/Echoer-64h0bb02goc.png

        The property :attr:`current_window` can be used to access the current
        window without processing it.

        >>> notes = echoer.current_window()
        >>> staff = abjad.Staff(notes)
        >>> abjad.show(staff)

        ..  docs::

            \new Staff
            {
                \time 4/4
                c'4
                \p
                d'4
                \pp
                e'4
                \ppp
                r4
            }

        ..  figure:: ../_images/Echoer-ruetb1tzhtn.png

    :attr:`process_on_first_call`:
        The very first call will output the input container without processing
        it. To disable this behaviour and apply the echoing process on the very
        first call, initialise the class with the keyword argument
        :attr:`process_on_first_call` set to ``True``.

        >>> container = abjad.Container(r"c'4\mf d'4\mp e'\p f'\pp")
        >>> echoer = auxjad.Echoer(container,
        ...                      process_on_first_call=True,
        ...                        )
        >>> notes = echoer()
        >>> staff = abjad.Staff(notes)
        >>> abjad.show(staff)

        ..  docs::

            \new Staff
            {
                \time 4/4
                c'4
                \mp
                d'4
                \p
                e'4
                \pp
                f'4
                \ppp
            }

        ..  figure:: ../_images/Echoer-rbnsf64kjoq.png

    :attr:`min_dynamic`:
        The threshold dynamic used to remove softer notes during each iteration
        of the process is set by :attr:`min_dynamic`.

        >>> container = abjad.Container(r"c'4\f d'4 e'4 f'4")
        >>> echoer = auxjad.Echoer(container,
        ...                        min_dynamic='mp',
        ...                        )
        >>> notes = echoer()
        >>> staff = abjad.Staff(notes)
        >>> abjad.show(staff)

        ..  docs::

            \new Staff
            {
                \time 4/4
                c'4
                \mf
                d'4
                \mp
                e'4
                \p
                f'4
                \pp
            }

        ..  figure:: ../_images/Echoer-dzjqv7lsdis.png

        >>> notes = echoer()
        >>> staff = abjad.Staff(notes)
        >>> abjad.show(staff)

        ..  docs::

            \new Staff
            {
                \time 4/4
                c'4
                \mp
                d'4
                \p
                r2
            }

        ..  figure:: ../_images/Echoer-f33perr6lfo.png

        >>> notes = echoer()
        >>> staff = abjad.Staff(notes)
        >>> abjad.show(staff)

        ..  docs::

            \new Staff
            {
                \time 4/4
                c'4
                \p
                r2.
            }

        ..  figure:: ../_images/Echoer-qxvj8lkfph.png

    Changing :attr:`min_dynamic` after initialisation:
        The property :attr:`min_dynamic` can also be changed after
        initialisation, as shown below.

        >>> container = abjad.Container(r"c'4\mf d'4\mp e'\p f'\pp")
        >>> echoer = auxjad.Echoer(container)
        >>> notes = echoer()
        >>> staff = abjad.Staff(notes)
        >>> abjad.show(staff)

        ..  docs::

            \new Staff
            {
                \time 4/4
                c'4
                \mf
                d'4
                \mp
                e'4
                \p
                f'4
                \pp
            }

        ..  figure:: ../_images/Echoer-3ldjnprohuo.png

        >>> notes = echoer()
        >>> staff = abjad.Staff(notes)
        >>> abjad.show(staff)

        ..  docs::

            \new Staff
            {
                \time 4/4
                c'4
                \mp
                d'4
                \p
                e'4
                \pp
                f'4
                \ppp
            }

        ..  figure:: ../_images/Echoer-wzeaqjgouz8.png

        >>> notes = echoer()
        >>> staff = abjad.Staff(notes)
        >>> abjad.show(staff)

        ..  docs::

            \new Staff
            {
                \time 4/4
                c'4
                \p
                d'4
                \pp
                e'4
                \ppp
                r4
            }

        ..  figure:: ../_images/Echoer-aqq1docvezb.png

        >>> echoer.min_dynamic = 'pp'
        >>> notes = echoer()
        >>> staff = abjad.Staff(notes)
        >>> abjad.show(staff)

        ..  docs::

            \new Staff
            {
                \time 4/4
                c'4
                \pp
                r2.
            }

        ..  figure:: ../_images/Echoer-3jt6kto85h1.png

        >>> notes = echoer()
        >>> staff = abjad.Staff(notes)
        >>> abjad.show(staff)

        ..  docs::

            \new Staff
            {
                \time 4/4
                R1
            }

        ..  figure:: ../_images/Echoer-2i22so5t5pf.png

    :attr:`include_empty_measures`:
        Set :attr:`include_empty_measures` to ``False`` to exclude empty
        measures at the end of the process (default is ``True``).

        >>> container = abjad.Container(r"c'4\p d'4 e' f'")
        >>> echoer = auxjad.Echoer(container,
        ...                        include_empty_measures=False,
        ...                        )
        >>> staff = abjad.Staff(echoer.output_all())
        >>> abjad.show(staff)

        ..  docs::

            \new Staff
            {
                \time 4/4
                c'4
                \p
                d'4
                e'4
                f'4
                c'4
                \pp
                d'4
                e'4
                f'4
                c'4
                \ppp
                d'4
                e'4
                f'4
            }

        ..  figure:: ../_images/Echoer-pg0bejhb7ke.png

    Using as iterator:
        The instances of this class can also be used as an iterator, which can
        then be used in a for loop to run through the whole process. Note that
        unlike the methods :meth:`output_n` and :meth:`output_all`, time
        signatures are added to each window returned by the echoer. Use the
        function |auxjad.mutate.remove_repeated_time_signatures()| to clean
        the output when using this class in this way.

        >>> container = abjad.Container(r"c'4\mf d'4\mp e'\p f'\pp")
        >>> echoer = auxjad.Echoer(container)
        >>> staff = abjad.Staff()
        >>> for window in echoer:
        ...     staff.append(window)
        >>> auxjad.mutate.remove_repeated_time_signatures(staff)
        >>> abjad.show(staff)

        ..  docs::

            \new Staff
            {
                \time 4/4
                c'4
                \mf
                d'4
                \mp
                e'4
                \p
                f'4
                \pp
                \time 4/4
                c'4
                \mp
                d'4
                \p
                e'4
                \pp
                f'4
                \ppp
                \time 4/4
                c'4
                \p
                d'4
                \pp
                e'4
                \ppp
                r4
                \time 4/4
                c'4
                \pp
                d'4
                \ppp
                r2
                \time 4/4
                c'4
                \ppp
                r2.
                \time 4/4
                R1
            }

        ..  figure:: ../_images/Echoer-qyve2exm08p.png

    Arguments and properties:
        This class can take many optional keyword arguments during its
        creation, besides :attr:`min_dynamic`. By default, calling the object
        for the first time will return the original container; set
        :attr:`process_on_first_call` to ``True`` and the echo process will be
        applied on the very first call. :attr:`max_steps` sets the maximum
        number of iterations of the echoing process that can be applied in a
        single call, ranging between ``1`` and the input value (default is also
        ``1``). :attr:`repetition_chance` sets the chance of a window repeating
        itself, from ``0.0`` to ``1.0`` (default is ``0.0``, i.e. no
        repetitions). :attr:`disable_rewrite_meter` disables the
        |abjad.Meter.rewrite_meter()| mutation which is applied to the
        container after every call, and :attr:`omit_time_signatures` will
        remove all time signatures from the output (both are ``False`` by
        default). Any measure filled with rests will be rewritten using a
        multi-measure rest; set the :attr:`use_multimeasure_rests` to ``False``
        to disable this behaviour. The properties :attr:`boundary_depth`,
        :attr:`maximum_dot_count`, and :attr:`rewrite_tuplets` are passed as
        arguments to |abjad.Meter.rewrite_meter()|, see its documentation for
        more information.

        >>> container = abjad.Container(r"c'4\mf d'4\mp e'\p f'\pp")
        >>> echoer = auxjad.Echoer(container,
        ...                        min_dynamic='p',
        ...                        max_steps=2,
        ...                        repetition_chance=0.7,
        ...                        disable_rewrite_meter=True,
        ...                        omit_time_signatures=True,
        ...                        use_multimeasure_rests=False,
        ...                        boundary_depth=0,
        ...                        maximum_dot_count=1,
        ...                        rewrite_tuplets=False,
        ...                        process_on_first_call=True,
        ...                        include_empty_measures=False,
        ...                        )
        >>> echoer.min_dynamic
        'p'
        >>> echoer.max_steps
        2
        >>> echoer.repetition_chance
        0.7
        >>> echoer.disable_rewrite_meter
        True
        >>> echoer.omit_time_signatures
        True
        >>> echoer.use_multimeasure_rests
        False
        >>> echoer.boundary_depth
        0
        >>> echoer.maximum_dot_count
        1
        >>> echoer.rewrite_tuplets
        False
        >>> echoer.process_on_first_call
        True
        >>> echoer.include_empty_measures
        False

        Use the properties below to change these values after initialisation.

        >>> echoer.min_dynamic = 'mp'
        >>> echoer.max_steps = 1
        >>> echoer.repetition_chance = 0.23
        >>> echoer.disable_rewrite_meter = False
        >>> echoer.omit_time_signatures = False
        >>> echoer.use_multimeasure_rests = True
        >>> echoer.boundary_depth = 1
        >>> echoer.maximum_dot_count = 2
        >>> echoer.rewrite_tuplets = True
        >>> echoer.process_on_first_call = False
        >>> echoer.include_empty_measures = True
        >>> echoer.min_dynamic
        'mp'
        >>> echoer.max_steps
        1
        >>> echoer.repetition_chance
        0.23
        >>> echoer.disable_rewrite_meter
        False
        >>> echoer.omit_time_signatures
        False
        >>> echoer.use_multimeasure_rests
        True
        >>> echoer.boundary_depth
        1
        >>> echoer.maximum_dot_count
        2
        >>> echoer.rewrite_tuplets
        True
        >>> echoer.process_on_first_call
        False
        >>> echoer.include_empty_measures
        True

    :attr:`contents`:
        Use the :attr:`contents` property to read as well as overwrite the
        contents of the echoer. Notice that the process will also be reset at
        that point.

        >>> container = abjad.Container(r"c'4\mf d'4\mp e'\p f'\pp")
        >>> echoer = auxjad.Echoer(container)
        >>> notes = echoer()
        >>> staff = abjad.Staff(notes)
        >>> abjad.show(staff)

        ..  docs::

            \new Staff
            {
                \time 4/4
                c'4
                \mf
                d'4
                \mp
                e'4
                \p
                f'4
                \pp
            }

        ..  figure:: ../_images/Echoer-nv5f76rv7f.png

        >>> notes = echoer()
        >>> staff = abjad.Staff(notes)
        >>> abjad.show(staff)

        ..  docs::

            \new Staff
            {
                \time 4/4
                c'4
                \mp
                d'4
                \p
                e'4
                \pp
                f'4
                \ppp
            }

        ..  figure:: ../_images/Echoer-6fr4wrb8god.png

        >>> echoer.contents = abjad.Container(r"c'16\f d'16 e'16 f'16 g'2.\p")
        >>> notes = echoer()
        >>> staff = abjad.Staff(notes)
        >>> abjad.show(staff)

        ..  docs::

            \new Staff
            {
                \time 4/4
                c'16
                \f
                d'16
                e'16
                f'16
                g'2.
                \p
            }

        ..  figure:: ../_images/Echoer-p8q5x8ti2d.png

        >>> notes = echoer()
        >>> staff = abjad.Staff(notes)
        >>> abjad.show(staff)

        ..  docs::

            \new Staff
            {
                \time 4/4
                c'16
                \mf
                d'16
                e'16
                f'16
                g'2.
                \pp
            }

        ..  figure:: ../_images/Echoer-mky4pulzf4i.png

    :meth:`output_all`:
        To run through the whole process and output it as a single container,
        use the method :meth:`output_all`.

        >>> container = abjad.Container(r"c'4\mf d'4\mp e'\p f'\pp")
        >>> echoer = auxjad.Echoer(container)
        >>> notes = echoer.output_all()
        >>> staff = abjad.Staff(notes)
        >>> abjad.show(staff)

        ..  docs::

            \new Staff
            {
                \time 4/4
                c'4
                \mf
                d'4
                \mp
                e'4
                \p
                f'4
                \pp
                c'4
                \mp
                d'4
                \p
                e'4
                \pp
                f'4
                \ppp
                c'4
                \p
                d'4
                \pp
                e'4
                \ppp
                r4
                c'4
                \pp
                d'4
                \ppp
                r2
                c'4
                r2.
                R1
            }

        ..  figure:: ../_images/Echoer-kiqwdhyx9vk.png

    :meth:`output_n`:
        To run through just part of the process and output it as a single
        container, use the method :meth:`output_n` and pass the number of
        iterations as argument.

        >>> container = abjad.Container(r"c'4\mf d'4\mp e'\p f'\pp")
        >>> echoer = auxjad.Echoer(container)
        >>> notes = echoer.output_n(3)
        >>> staff = abjad.Staff(notes)
        >>> abjad.show(staff)

        ..  docs::

            \new Staff
            {
                \time 4/4
                c'4
                \mf
                d'4
                \mp
                e'4
                \p
                f'4
                \pp
                c'4
                \mp
                d'4
                \p
                e'4
                \pp
                f'4
                \ppp
                c'4
                \p
                d'4
                \pp
                e'4
                \ppp
                r4
            }

        ..  figure:: ../_images/Echoer-6mqxj9b5f13.png

    Chords and rests:
        This class also support chords and rest in :attr:`contents`.

        >>> container = abjad.Container(
        ...     r"<c' g' e'>4\p ~ <c' g' e'>16 r8. r8 <d' f' a'>4.\mp"
        ... )
        >>> echoer = auxjad.Echoer(container)
        >>> staff = abjad.Staff(echoer.output_all())
        >>> abjad.show(staff)

        ..  docs::

            \new Staff
            {
                \time 4/4
                <c' e' g'>4
                \p
                ~
                <c' e' g'>16
                r8.
                r8
                <d' f' a'>4.
                \mp
                <c' e' g'>4
                \pp
                ~
                <c' e' g'>16
                r8.
                r8
                <d' f' a'>4.
                \p
                <c' e' g'>4
                \ppp
                ~
                <c' e' g'>16
                r8.
                r8
                <d' f' a'>4.
                \pp
                r2
                r8
                <d' f' a'>4.
                \ppp
                R1
            }

        ..  figure:: ../_images/Echoer-wjphyrz750d.png

    :func:`len()`:
        The function :func:`len()` returns the total number of notes in
        :attr:`contents`.

        >>> abjad.Container(r"c'4\mf d'4\mp e'\p f'\pp")
        >>> echoer = auxjad.Echoer(container)
        >>> len(echoer)
        4
        >>> container = abjad.Container(
        ...     r"c'4\mf ~ c'8 d'8\mp e'4\p ~ e'8 f'8\pp"
        ... )
        >>> echoer = auxjad.Echoer(container)
        >>> len(echoer)
        4
        >>> container = abjad.Container(
        ...     r"c'4\mf ~ c'16 r16 d'8\mp e'4\p ~ e'8 f'16\pp r16"
        ... )
        >>> echoer = auxjad.Echoer(container)
        >>> len(echoer)
        4

        Note that each chord will count as a single unit.

        >>> container = abjad.Container(r"<c' e' g'>2\f <d' f'>2\p")
        >>> echoer = auxjad.Echoer(container)
        >>> len(echoer)
        2
        >>> container = abjad.Container(
        ...     r"<c' e' g'>4\f ~ <c' e' g'>16 r8. <d' f'>2\p"
        ... )
        >>> echoer = auxjad.Echoer(container)
        >>> len(echoer)
        2
        >>> container = abjad.Container(
        ...     r"<c' e' g'>4\f d'4\mf <e' g' b'>4\p r4"
        ... )
        >>> echoer = auxjad.Echoer(container)
        >>> len(echoer)
        3

    :attr:`max_steps`:
        Setting the keyword argument :attr:`max_steps` to a value larger than
        ``1`` will result in a random number of steps (between ``1`` and
        :attr:`max_steps`) being applied at each call.

        >>> container = abjad.Container(r"c'2\fff d'2\mf")
        >>> echoer = auxjad.Echoer(container,
        ...                        max_steps=3,
        ...                        )
        >>> notes = echoer.output_n(3)
        >>> staff = abjad.Staff(notes)
        >>> abjad.show(staff)

        ..  docs::

            \new Staff
            {
                \time 4/4
                c'2
                \fff
                d'2
                \mf
                c'2
                \f
                d'2
                \p
                c'2
                \mf
                d'2
                \pp
            }

        ..  figure:: ../_images/Echoer-72wpb0iqtes.png

    :attr:`repetition_chance`:
        Use :attr:`repetition_chance` to set the chance of a measure repeating
        itself, ranging from ``0.0`` to ``1.0`` (default is ``0.0``,
        i.e. no repetitions).

        >>> container = abjad.Container(r"c'4.\f d'8\p e'4..\mf f'16\mp")
        >>> echoer = auxjad.Echoer(container,
        ...                        repetition_chance=0.5,
        ...                        )
        >>> notes = echoer.output_n(5)
        >>> staff = abjad.Staff(notes)
        >>> abjad.show(staff)

        ..  docs::

            \new Staff
            {
                \time 4/4
                c'4.
                \f
                d'8
                \p
                e'4..
                \mf
                f'16
                \mp
                c'4.
                \f
                d'8
                \p
                e'4..
                \mf
                f'16
                \mp
                c'4.
                \mf
                d'8
                \pp
                e'4..
                \mp
                f'16
                \p
                c'4.
                \mp
                d'8
                \ppp
                e'4..
                \p
                f'16
                \pp
                c'4.
                \mp
                d'8
                \ppp
                e'4..
                \p
                f'16
                \pp
            }

        ..  figure:: ../_images/Echoer-1t2yh8imiu8.png

    :meth:`reset`:
        Use the method :meth:`reset` to reset the process.

        >>> container = abjad.Container(r"c'4\mf d'4\mp e'\p f'\pp")
        >>> echoer = auxjad.Echoer(container)
        >>> notes = echoer()
        >>> abjad.show(staff)

        ..  docs::

            \new Staff
            {
                \time 4/4
                c'4
                \mf
                d'4
                \mp
                e'4
                \p
                f'4
                \pp
            }

        ..  figure:: ../_images/Echoer-Is3abUvpyU.png

        >>> notes = echoer()
        >>> abjad.show(staff)

        ..  docs::

            \new Staff
            {
                \time 4/4
                c'4
                \mp
                d'4
                \p
                e'4
                \pp
                f'4
                \ppp
            }

        ..  figure:: ../_images/Echoer-gNgOHWmr1T.png

        >>> echoer.reset()
        >>> notes = echoer()
        >>> abjad.show(staff)

        ..  docs::

            \new Staff
            {
                \time 4/4
                c'4
                \mf
                d'4
                \mp
                e'4
                \p
                f'4
                \pp
            }

        ..  figure:: ../_images/Echoer-ioSV8IbIIL.png

    :attr:`use_multimeasure_rests` and :attr:`disable_rewrite_meter`:
        By default, all rests in a measure filled only with rests will be
        converted into a multi-measure rest. Set :attr:`use_multimeasure_rests`
        to ``False`` to disable this. Also, by default, all output is mutated
        through |abjad.Meter.rewrite_meter()|. To disable it, set
        :attr:`disable_rewrite_meter` to ``True``.

        >>> container = abjad.Container(r"c'4\p ~ c'16 d'8.\mp ~ d'2")
        >>> echoer = auxjad.Echoer(container,
        ...                        disable_rewrite_meter=True,
        ...                        use_multimeasure_rests=False,
        ...                        )
        >>> notes = echoer.output_all()
        >>> staff = abjad.Staff(notes)
        >>> abjad.show(staff)

        ..  docs::

            \new Staff
            {
                \time 4/4
                c'4
                \p
                ~
                c'16
                d'8.
                \mp
                ~
                d'2
                c'4
                \pp
                ~
                c'16
                d'8.
                \p
                ~
                d'2
                c'4
                \ppp
                ~
                c'16
                d'8.
                \pp
                ~
                d'2
                r4
                r16
                d'8.
                \ppp
                ~
                d'2
                r4
                r16
                r8.
                r2
            }

        ..  figure:: ../_images/Echoer-lax06gkb3ap.png

    :attr:`omit_time_signatures`:
        To disable time signatures altogether, initialise this class with the
        keyword argument :attr:`omit_time_signatures` set to ``True`` (default
        is ``False``), or use the :attr:`omit_time_signatures` property after
        initialisation.

        >>> container = abjad.Container(
        ...     r"\time 2/4 c'4\mf d'4 \time 3/4 e'4\p f'4 g'4"
        ... )
        >>> echoer = auxjad.Echoer(container,
        ...                        omit_time_signatures=True,
        ...                        )
        >>> notes = echoer.output_n(3)
        >>> staff = abjad.Staff(notes)
        >>> abjad.show(staff)

        ..  docs::

            \new Staff
            {
                c'4
                \mf
                d'4
                e'4
                \p
                f'4
                g'4
                c'4
                \mp
                d'4
                e'4
                \pp
                f'4
                g'4
                c'4
                \p
                d'4
                e'4
                \ppp
                f'4
                g'4
            }

        ..  figure:: ../_images/Echoer-pcq5ecwz7ff.png

    ..  tip::

        All methods that return an |abjad.Selection| will add an initial time
        signature to it. The :meth:`output_n` and :meth:`output_all` methods
        automatically remove repeated time signatures. When joining selections
        output by multiple method calls, use
        |auxjad.mutate.remove_repeated_time_signatures()| on the whole
        container after fusing the selections to remove any unecessary time
        signature changes.

    Tweaking |abjad.Meter.rewrite_meter()|:
        This function uses the default logical tie splitting algorithm from
        |abjad.Meter.rewrite_meter()|.

        >>> container = abjad.Container(r"c'4.\mf d'8 e'2")
        >>> echoer = auxjad.Echoer(container)
        >>> notes = echoer()
        >>> staff = abjad.Staff(notes)
        >>> abjad.show(staff)

        ..  docs::

            \new Staff
            {
                \time 4/4
                c'4.
                \mf
                d'8
                e'2
            }

        ..  figure:: ../_images/Echoer-93hcv2prkua.png

        Set :attr:`boundary_depth` to a different number to change its
        behaviour.

        >>> echoer = auxjad.Echoer(container,
        ...                        boundary_depth=1,
        ...                        )
        >>> notes = echoer()
        >>> staff = abjad.Staff(notes)
        >>> abjad.show(staff)

        ..  docs::

            \new Staff
            {
                \time 4/4
                c'4
                \mf
                ~
                c'8
                d'8
                e'2
            }

        ..  figure:: ../_images/Echoer-cq661zyctf.png

        Other arguments available for tweaking the output of
        |abjad.Meter.rewrite_meter()| are :attr:`maximum_dot_count` and
        :attr:`rewrite_tuplets`, which work exactly as the identically named
        arguments of |abjad.Meter.rewrite_meter()|.

        This class also accepts the arguments ``fuse_across_groups_of_beats``,
        ``fuse_quadruple_meter``, ``fuse_triple_meter``, and
        ``extract_trivial_tuplets``, which are passed on to
        |auxjad.mutate.prettify_rewrite_meter()| (the latter can be disabled
        by setting ``prettify_rewrite_meter`` to ``False``). See the
        documentation of this function for more details on these arguments.

    Indicators:
        This class can handle dynamics and articulations.

        >>> container = abjad.Container(
        ...     r"\time 3/4 c'8.->\mf d'16 ~ d'4 e'8..--\p f'32-.\f"
        ... )
        >>> echoer = auxjad.Echoer(container)
        >>> notes = echoer.output_all()
        >>> staff = abjad.Staff(notes)
        >>> abjad.show(staff)

        ..  docs::

            \new Staff
            {
                \time 3/4
                c'8.
                \mf
                - \accent
                d'16
                ~
                d'4
                e'8..
                \p
                - \tenuto
                f'32
                \f
                - \staccato
                c'8.
                \mp
                - \accent
                d'16
                ~
                d'4
                e'8..
                \pp
                - \tenuto
                f'32
                \mf
                - \staccato
                c'8.
                \p
                - \accent
                d'16
                ~
                d'4
                e'8..
                \ppp
                - \tenuto
                f'32
                \mp
                - \staccato
                c'8.
                \pp
                - \accent
                d'16
                ~
                d'4
                r8..
                f'32
                \p
                - \staccato
                c'8.
                \ppp
                - \accent
                d'16
                ~
                d'4
                r8..
                f'32
                \pp
                - \staccato
                r2
                r8..
                f'32
                \ppp
                - \staccato
                R1 * 3/4
            }

        ..  figure:: ../_images/Echoer-ox08wd3ljps.png

    Slurs and hairpins:
        Slurs and hairpins are also supported. Slurs are split when rests
        appear in the middle of a slurred phrase, while hairpins are shortened
        and adjusted as required.

        >>> container = abjad.Container(
        ...     r"\times 2/3 {c'2(\p\< d'2 e'2\ff} f'4\mf\> g'2 a'4\mp)"
        ... )
        >>> echoer = auxjad.Echoer(container)
        >>> notes = echoer.output_n(5)
        >>> staff = abjad.Staff(notes)
        >>> abjad.show(staff)

        ..  docs::

            \new Staff
            {
                \times 2/3
                {
                    \time 4/4
                    c'2
                    \p
                    \<
                    (
                    d'2
                    e'2
                    \ff
                }
                f'4
                \mf
                \>
                g'2
                a'4
                \mp
                )
                \times 2/3
                {
                    c'2
                    \pp
                    \<
                    (
                    d'2
                    e'2
                    \f
                }
                f'4
                \mp
                \>
                g'2
                a'4
                \p
                )
                \times 2/3
                {
                    c'2
                    \ppp
                    \<
                    (
                    d'2
                    e'2
                    \mf
                }
                f'4
                \p
                \>
                g'2
                a'4
                \pp
                )
                \times 2/3
                {
                    r1
                    e'2
                    \mp
                    (
                }
                f'4
                \pp
                \>
                g'2
                a'4
                \ppp
                )
                \times 2/3
                {
                    r1
                    e'2
                    \p
                    (
                }
                f'4
                \ppp
                \>
                g'2
                )
                r4
                \!
            }

        ..  figure:: ../_images/Echoer-jprjps4zxej.png

    ..  tip::

        The functions |auxjad.mutate.remove_repeated_dynamics()| and
        |auxjad.mutate.reposition_clefs()| can be used to clean the output
        and remove repeated dynamics and unnecessary clef changes.

    ..  warning::

        Do note that some elements that span multiple notes (such as ottava
        indicators, manual beams, etc.) can become problematic when notes
        containing them are split into two. As a rule of thumb, it is always
        better to attach those to the music after the echoing process has
        ended.

    Tuplets:
        This class can handle tuplets.

        >>> container = abjad.Container(
        ...     r"\times 2/3 {c'8\ppp d'8\mp e'8} d'2.\pp"
        ... )
        >>> echoer = auxjad.Echoer(container)
        >>> notes = echoer.output_all()
        >>> staff = abjad.Staff(notes)
        >>> abjad.show(staff)

        ..  docs::

            \new Staff
            {
                \times 2/3
                {
                    \time 4/4
                    c'8
                    \ppp
                    d'8
                    \mp
                    e'8
                }
                d'2.
                \pp
                \times 2/3
                {
                    r8
                    d'8
                    \p
                    e'8
                }
                d'2.
                \ppp
                \times 2/3
                {
                    r8
                    d'8
                    \pp
                    e'8
                }
                r2.
                \times 2/3
                {
                    r8
                    d'8
                    \ppp
                    e'8
                }
                r2.
                R1
            }

        ..  figure:: ../_images/Echoer-888tqk73kw3.png

    Time signature changes:
        This class can handle time signature changes.

        >>> container = abjad.Container(
        ...     r"\time 3/8 c'4.\pp \time 2/4 d'2\ff \time 3/8 e'4.\mp"
        ... )
        >>> echoer = auxjad.Echoer(container)
        >>> notes = echoer.output_n()
        >>> staff = abjad.Staff(notes)
        >>> abjad.show(staff)

        ..  docs::

            \new Staff
            {
                \time 3/8
                c'4.
                \pp
                \time 2/4
                d'2
                \ff
                \time 3/8
                e'4.
                \mp
                c'4.
                \ppp
                \time 2/4
                d'2
                \f
                \time 3/8
                e'4.
                \p
                R1 * 3/8
                \time 2/4
                d'2
                \mf
                \time 3/8
                e'4.
                \pp
            }

        ..  figure:: ../_images/Echoer-lkhKFVuUgx.png
    """

    ### CLASS VARIABLES ###

    __slots__ = (
        "_contents",
        "_current_window",
        "_min_dynamic",
        "_max_steps",
        "_disable_rewrite_meter",
        "_mask",
        "_is_first_window",
        "_time_signatures",
        "_omit_time_signatures",
        "_use_multimeasure_rests",
        "_boundary_depth",
        "_maximum_dot_count",
        "_rewrite_tuplets",
        "_process_on_first_call",
        "_include_empty_measures",
        "_repetition_chance",
        "_prettify_rewrite_meter",
        "_extract_trivial_tuplets",
        "_fuse_across_groups_of_beats",
        "_fuse_quadruple_meter",
        "_fuse_triple_meter",
    )

    _dynamic_name_to_dynamic_ordinal = {
        "ppppp": -5,
        "pppp": -4,
        "ppp": -3,
        "pp": -2,
        "p": -1,
        "mp": 0,
        "mf": 1,
        "f": 2,
        "ff": 3,
        "fff": 4,
        "ffff": 5,
        "fffff": 6,
        "fp": -1,
        "sf": 2,
        "sff": 3,
        "sfp": -1,
        "sfpp": -2,
        "sffp": -1,
        "sffpp": -2,
        "sfz": 2,
        "sp": -1,
        "spp": -2,
        "rfz": 2,
    }

    _dynamic_ordinal_to_dynamic_name = {
        -5: "ppppp",
        -4: "pppp",
        -3: "ppp",
        -2: "pp",
        -1: "p",
        0: "mp",
        1: "mf",
        2: "f",
        3: "ff",
        4: "fff",
        5: "ffff",
        6: "fffff",
    }

    ### INITIALISER ###

    def __init__(
        self,
        contents: abjad.Container,
        *,
        min_dynamic: Union[abjad.Dynamic, str] = "ppp",
        max_steps: int = 1,
        repetition_chance: float = 0.0,
        process_on_first_call: bool = False,
        disable_rewrite_meter: bool = False,
        omit_time_signatures: bool = False,
        use_multimeasure_rests: bool = True,
        boundary_depth: Optional[int] = None,
        maximum_dot_count: Optional[int] = None,
        rewrite_tuplets: bool = True,
        include_empty_measures: bool = True,
        prettify_rewrite_meter: bool = True,
        extract_trivial_tuplets: bool = True,
        fuse_across_groups_of_beats: bool = True,
        fuse_quadruple_meter: bool = True,
        fuse_triple_meter: bool = True,
    ) -> None:
        self.min_dynamic = min_dynamic
        self.max_steps = max_steps
        self.contents = contents
        self.disable_rewrite_meter = disable_rewrite_meter
        self.omit_time_signatures = omit_time_signatures
        self.use_multimeasure_rests = use_multimeasure_rests
        self.boundary_depth = boundary_depth
        self.maximum_dot_count = maximum_dot_count
        self.rewrite_tuplets = rewrite_tuplets
        self.prettify_rewrite_meter = prettify_rewrite_meter
        self.extract_trivial_tuplets = extract_trivial_tuplets
        self.fuse_across_groups_of_beats = fuse_across_groups_of_beats
        self.fuse_quadruple_meter = fuse_quadruple_meter
        self.fuse_triple_meter = fuse_triple_meter
        self.process_on_first_call = process_on_first_call
        self.include_empty_measures = include_empty_measures
        self.repetition_chance = repetition_chance
        self._is_first_window = True

    ### SPECIAL METHODS ###

    def __repr__(self) -> str:
        r"""Returns interpreter representation of  :attr:`contents`."""
        return abjad.lilypond(self._contents)

    def __len__(self) -> int:
        r"""Returns the number of logical ties of :attr:`contents`."""
        logical_ties = abjad.select(self._contents).logical_ties(pitched=True)
        return len(logical_ties)

    def __call__(self) -> abjad.Selection:
        r"""Calls the echo process for one iteration, returning an
        |abjad.Selection|.
        """
        if self._repetition_chance == 0.0 or random.random() > self._repetition_chance:
            if not self._is_first_window or self._process_on_first_call:
                self._soften_mask()
        self._mask_to_selection()
        return self.current_window

    def __next__(self) -> abjad.Selection:
        r"""Calls the echoing process for one iteration, returning an
        |abjad.Selection|.
        """
        if self._done:
            raise StopIteration
        return self.__call__()

    def __iter__(self) -> None:
        r"""Returns an iterator, allowing instances to be used as iterators."""
        return self

    ### PUBLIC METHODS ###

    def output_all(self) -> abjad.Selection:
        r"""Goes through the whole echoing process and outputs a single
        |abjad.Selection|.
        """
        dummy_container = abjad.Container()
        while True:
            dummy_container.append(self.__call__())
            if self._done:
                break
        mutate.remove_repeated_time_signatures(dummy_container[:])
        mutate.remove_repeated_dynamics(dummy_container[:])
        output = dummy_container[:]
        dummy_container[:] = []
        return output

    def output_n(
        self,
        n: int,
    ) -> abjad.Selection:
        r"""Goes through ``n`` iterations of the echoing process and outputs a
        single |abjad.Selection|.
        """
        if not isinstance(n, int):
            raise TypeError("first positional argument must be 'int'")
        if n < 1:
            raise ValueError("first positional argument must be a positive 'int'")
        dummy_container = abjad.Container()
        for _ in range(n):
            dummy_container.append(self.__call__())
        mutate.remove_repeated_time_signatures(dummy_container[:])
        mutate.remove_repeated_dynamics(dummy_container[:])
        output = dummy_container[:]
        dummy_container[:] = []
        return output

    def reset(self) -> None:
        r"""Resets the process, regenerating the mask."""
        self._is_first_window = True
        self._get_mask()

    ### PRIVATE METHODS ###

    def _mask_to_selection(self) -> None:
        r"""Applies the mask to :attr:`contents`."""
        dummy_container = abjad.mutate.copy(self._contents)
        if not self._is_first_window or self._process_on_first_call:
            logical_ties = abjad.select(dummy_container).logical_ties(
                pitched=True,
            )
            dyn_list = self._mask_to_dyn_list()
            previous_dyn = None
            for logical_tie, new_dyn in zip(logical_ties, dyn_list):
                current_dyn = abjad.get.indicator(
                    logical_tie.head,
                    abjad.Dynamic,
                )
                if current_dyn is not None:
                    abjad.detach(current_dyn, logical_tie.head)
                if new_dyn is not None:
                    if new_dyn != previous_dyn:
                        abjad.attach(new_dyn, logical_tie.head)
                else:
                    self._convert_pitched_logical_tie_to_rest(logical_tie)
                previous_dyn = new_dyn
        # handling dynamics and slurs and empty tuplets
        mutate.reposition_dynamics(dummy_container[:])
        mutate.reposition_slurs(dummy_container[:])
        mutate.extract_trivial_tuplets(dummy_container[:])
        # applying rewrite meter
        if not self._disable_rewrite_meter:
            mutate.auto_rewrite_meter(
                dummy_container,
                boundary_depth=self._boundary_depth,
                maximum_dot_count=self._maximum_dot_count,
                rewrite_tuplets=self._rewrite_tuplets,
                prettify_rewrite_meter=self._prettify_rewrite_meter,
                extract_trivial_tuplets=self._extract_trivial_tuplets,
                fuse_across_groups_of_beats=self._fuse_across_groups_of_beats,
                fuse_quadruple_meter=self._fuse_quadruple_meter,
                fuse_triple_meter=self._fuse_triple_meter,
            )
        if self._use_multimeasure_rests:
            mutate.rests_to_multimeasure_rest(dummy_container[:])
        # output
        self._current_window = dummy_container[:]
        dummy_container[:] = []
        self._is_first_window = False

    def _get_lilypond_format(self) -> str:
        r"""Returns interpreter representation of  :attr:`contents`."""
        return self.__repr__()

    def _get_mask(self) -> None:
        r"""Creates a mask of dynamic ordinals from :attr:`contents`."""
        logical_ties = abjad.select(self._contents).logical_ties(pitched=True)
        self._mask = []
        previous_dyn = None
        for logical_tie in logical_ties:
            dyn = abjad.get.indicator(logical_tie.head, abjad.Dynamic)
            if dyn is not None:
                try:
                    dyn = self._dynamic_name_to_dynamic_ordinal[dyn.name]
                except KeyError as exc:
                    raise ValueError(f"the dynamic marking '{dyn.name}' is not supported") from exc
            else:
                dyn = previous_dyn
            self._mask.append(dyn)
            previous_dyn = dyn
        if self._mask[0] is None:
            raise RuntimeError("first note of 'contents' must have a dynamic")

    def _mask_to_dyn_list(self) -> list[str]:
        r"""Converts the numerical mask into a list of dynamic strings."""
        dyn_list = []
        for dyn_number in self._mask:
            if dyn_number is not None:
                dyn = self._dynamic_ordinal_to_dynamic_name[dyn_number]
                dyn_list.append(abjad.Dynamic(dyn))
            else:
                dyn_list.append(None)
        return dyn_list

    def _soften_mask(self) -> list[Union[int, None]]:
        r"""Lowers the dynamics of the mask by one level."""
        for n in range(random.randint(1, self._max_steps)):
            if any(item is not None for item in self._mask):
                self._mask = [
                    self._soften_dynamic(
                        item,
                        self._min_dynamic_number,
                    )
                    for item in self._mask
                ]
            elif n == 0:
                raise RuntimeError("'current_window' is already empty")

    @staticmethod
    def _soften_dynamic(
        item: Union[int, None],
        min_dyn_number: int,
    ) -> Union[int, None]:
        r"""Lowers a numerical dynamic level by one, setting it to ``None`` if
        dynamic below a given threshold.
        """
        if item is None:
            return None
        item -= 1
        if item < min_dyn_number:
            return None
        return item

    @staticmethod
    def _convert_pitched_logical_tie_to_rest(
        logical_tie: abjad.LogicalTie,
    ) -> None:
        r"""Converts all leaves of a pitched logical tie into rests."""
        indicators_tuple = (
            abjad.BarLine,
            abjad.Clef,
            abjad.Dynamic,
            abjad.Fermata,
            abjad.KeySignature,
            abjad.LilyPondLiteral,
            abjad.MetronomeMark,
            abjad.Ottava,
            abjad.RehearsalMark,
            abjad.Repeat,
            abjad.StaffChange,
            abjad.StartHairpin,
            abjad.StartMarkup,
            abjad.StartPhrasingSlur,
            abjad.StartSlur,
            abjad.StartTextSpan,
            abjad.StopHairpin,
            abjad.StopPhrasingSlur,
            abjad.StopSlur,
            abjad.StopTextSpan,
            abjad.TimeSignature,
        )
        for leaf in logical_tie:
            rest = abjad.Rest(leaf.written_duration)
            for indicator in abjad.get.indicators(leaf):
                if isinstance(indicator, indicators_tuple):
                    abjad.attach(indicator, rest)
            abjad.mutate.replace(leaf, rest)

    @staticmethod
    def _remove_all_time_signatures(
        container: abjad.Container,
    ) -> None:
        r"""Removes all time signatures of an |abjad.Container|."""
        for leaf in abjad.select(container).leaves():
            if abjad.get.effective(leaf, abjad.TimeSignature):
                abjad.detach(abjad.TimeSignature, leaf)

    ### PUBLIC PROPERTIES ###

    @property
    def contents(self) -> abjad.Container:
        r"""The |abjad.Container| to be faded."""
        return abjad.mutate.copy(self._contents)

    @contents.setter
    def contents(
        self,
        contents: abjad.Container,
    ) -> None:
        if not isinstance(contents, abjad.Container):
            raise TypeError("'contents' must be 'abjad.Container' or child class")
        if not abjad.select(contents).leaves().are_contiguous_logical_voice():
            raise ValueError("'contents' must be contiguous logical voice")
        if isinstance(contents, abjad.Score):
            self._contents = abjad.mutate.copy(contents[0])
        elif isinstance(contents, abjad.Tuplet):
            self._contents = abjad.Container([abjad.mutate.copy(contents)])
        else:
            self._contents = abjad.mutate.copy(contents)
        time_signatures = get.time_signature_list(
            self._contents,
            do_not_use_none=True,
        )
        mutate.enforce_time_signature(self._contents, time_signatures)
        dummy_container = abjad.mutate.copy(contents)
        self._current_window = dummy_container[:]
        dummy_container[:] = []
        self.reset()
        self._is_first_window = True

    @property
    def current_window(self) -> abjad.Selection:
        r"""Read-only property, returns the previously output selection."""
        current_window = abjad.mutate.copy(self._current_window)
        if self._omit_time_signatures:
            self._remove_all_time_signatures(current_window)
        return current_window

    @property
    def min_dynamic(self) -> str:
        r"""The minimum dynamic below which notes are removed."""
        return self._min_dynamic

    @min_dynamic.setter
    def min_dynamic(
        self,
        min_dynamic: Union[abjad.Dynamic, str],
    ) -> None:
        if not isinstance(min_dynamic, (abjad.Dynamic, str)):
            raise TypeError("'min_dynamic' must be 'abjad.Dynamic' or 'str'")
        if isinstance(min_dynamic, abjad.Dynamic):
            self._min_dynamic = min_dynamic.name
        else:
            self._min_dynamic = min_dynamic

    @property
    def max_steps(self) -> int:
        r"""The maximum number of steps per operation."""
        return self._max_steps

    @max_steps.setter
    def max_steps(
        self,
        max_steps: int,
    ) -> None:
        if not isinstance(max_steps, int):
            raise TypeError("'max_steps' must be 'int'")
        if max_steps < 1:
            raise ValueError("'max_steps' must be greater than zero")
        self._max_steps = max_steps

    @property
    def disable_rewrite_meter(self) -> bool:
        r"""When ``True``, the durations of the notes in the output will not be
        rewritten by the |abjad.Meter.rewrite_meter()| mutation.
        """
        return self._disable_rewrite_meter

    @disable_rewrite_meter.setter
    def disable_rewrite_meter(
        self,
        disable_rewrite_meter: bool,
    ) -> None:
        if not isinstance(disable_rewrite_meter, bool):
            raise TypeError("'disable_rewrite_meter' must be 'bool'")
        self._disable_rewrite_meter = disable_rewrite_meter

    @property
    def omit_time_signatures(self) -> bool:
        r"""When ``True``, all time signatures will be omitted from the
        output.
        """
        return self._omit_time_signatures

    @omit_time_signatures.setter
    def omit_time_signatures(
        self,
        omit_time_signatures: bool,
    ) -> None:
        if not isinstance(omit_time_signatures, bool):
            raise TypeError("'omit_time_signatures' must be 'bool'")
        self._omit_time_signatures = omit_time_signatures

    @property
    def use_multimeasure_rests(self) -> bool:
        r"""When ``True``, multi-measure rests will be used for silent
        measures.
        """
        return self._use_multimeasure_rests

    @use_multimeasure_rests.setter
    def use_multimeasure_rests(
        self,
        use_multimeasure_rests: bool,
    ) -> None:
        if not isinstance(use_multimeasure_rests, bool):
            raise TypeError("'use_multimeasure_rests' must be 'bool'")
        self._use_multimeasure_rests = use_multimeasure_rests

    @property
    def boundary_depth(self) -> Union[int, None]:
        r"""Sets the argument ``boundary_depth`` of
        |abjad.Meter.rewrite_meter()|.
        """
        return self._boundary_depth

    @boundary_depth.setter
    def boundary_depth(
        self,
        boundary_depth: Optional[int],
    ) -> None:
        if boundary_depth is not None:
            if not isinstance(boundary_depth, int):
                raise TypeError("'boundary_depth' must be 'int'")
        self._boundary_depth = boundary_depth

    @property
    def maximum_dot_count(self) -> Union[int, None]:
        r"""Sets the argument ``maximum_dot_count`` of
        |abjad.Meter.rewrite_meter()|.
        """
        return self._maximum_dot_count

    @maximum_dot_count.setter
    def maximum_dot_count(
        self,
        maximum_dot_count: Optional[int],
    ) -> None:
        if maximum_dot_count is not None:
            if not isinstance(maximum_dot_count, int):
                raise TypeError("'maximum_dot_count' must be 'int'")
        self._maximum_dot_count = maximum_dot_count

    @property
    def rewrite_tuplets(self) -> bool:
        r"""Sets the argument ``rewrite_tuplets`` of
        |abjad.Meter.rewrite_meter()|.
        """
        return self._rewrite_tuplets

    @rewrite_tuplets.setter
    def rewrite_tuplets(
        self,
        rewrite_tuplets: bool,
    ) -> None:
        if not isinstance(rewrite_tuplets, bool):
            raise TypeError("'rewrite_tuplets' must be 'bool'")
        self._rewrite_tuplets = rewrite_tuplets

    @property
    def prettify_rewrite_meter(self) -> bool:
        r"""Used to enable or disable the mutation
        |auxjad.mutate.prettify_rewrite_meter()| (default ``True``).
        """
        return self._prettify_rewrite_meter

    @prettify_rewrite_meter.setter
    def prettify_rewrite_meter(
        self,
        prettify_rewrite_meter: bool,
    ) -> None:
        if not isinstance(prettify_rewrite_meter, bool):
            raise TypeError("'prettify_rewrite_meter' must be 'bool'")
        self._prettify_rewrite_meter = prettify_rewrite_meter

    @property
    def extract_trivial_tuplets(self) -> bool:
        r"""Sets the argument ``extract_trivial_tuplets`` of
        |auxjad.mutate.prettify_rewrite_meter()|.
        """
        return self._extract_trivial_tuplets

    @extract_trivial_tuplets.setter
    def extract_trivial_tuplets(
        self,
        extract_trivial_tuplets: bool,
    ) -> None:
        if not isinstance(extract_trivial_tuplets, bool):
            raise TypeError("'extract_trivial_tuplets' must be 'bool'")
        self._extract_trivial_tuplets = extract_trivial_tuplets

    @property
    def fuse_across_groups_of_beats(self) -> bool:
        r"""Sets the argument ``fuse_across_groups_of_beats`` of
        |auxjad.mutate.prettify_rewrite_meter()|.
        """
        return self._fuse_across_groups_of_beats

    @fuse_across_groups_of_beats.setter
    def fuse_across_groups_of_beats(
        self,
        fuse_across_groups_of_beats: bool,
    ) -> None:
        if not isinstance(fuse_across_groups_of_beats, bool):
            raise TypeError("'fuse_across_groups_of_beats' must be 'bool'")
        self._fuse_across_groups_of_beats = fuse_across_groups_of_beats

    @property
    def fuse_quadruple_meter(self) -> bool:
        r"""Sets the argument ``fuse_quadruple_meter`` of
        |auxjad.mutate.prettify_rewrite_meter()|.
        """
        return self._fuse_quadruple_meter

    @fuse_quadruple_meter.setter
    def fuse_quadruple_meter(
        self,
        fuse_quadruple_meter: bool,
    ) -> None:
        if not isinstance(fuse_quadruple_meter, bool):
            raise TypeError("'fuse_quadruple_meter' must be 'bool'")
        self._fuse_quadruple_meter = fuse_quadruple_meter

    @property
    def fuse_triple_meter(self) -> bool:
        r"""Sets the argument ``fuse_triple_meter`` of
        |auxjad.mutate.prettify_rewrite_meter()|.
        """
        return self._fuse_triple_meter

    @fuse_triple_meter.setter
    def fuse_triple_meter(
        self,
        fuse_triple_meter: bool,
    ) -> None:
        if not isinstance(fuse_triple_meter, bool):
            raise TypeError("'fuse_triple_meter' must be 'bool'")
        self._fuse_triple_meter = fuse_triple_meter

    @property
    def process_on_first_call(self) -> bool:
        r"""If ``True`` then :attr:`contents` will be processed in the very
        first call.
        """
        return self._process_on_first_call

    @process_on_first_call.setter
    def process_on_first_call(
        self,
        process_on_first_call: bool,
    ) -> None:
        if not isinstance(process_on_first_call, bool):
            raise TypeError("'process_on_first_call' must be 'bool'")
        self._process_on_first_call = process_on_first_call

    @property
    def include_empty_measures(self) -> bool:
        r"""If ``True`` then an initial or final empty measures will be used,
        otherwise the process starts/ends with a single logical tie.
        """
        return self._include_empty_measures

    @include_empty_measures.setter
    def include_empty_measures(
        self,
        include_empty_measures: bool,
    ) -> None:
        if not isinstance(include_empty_measures, bool):
            raise TypeError("'include_empty_measures' must be 'bool'")
        self._include_empty_measures = include_empty_measures

    @property
    def repetition_chance(self) -> float:
        r"""The chance of not processing :attr:`contents` on a call, thus
        repeating the previous output.
        """
        return self._repetition_chance

    @repetition_chance.setter
    def repetition_chance(
        self,
        repetition_chance: float,
    ) -> None:
        if not isinstance(repetition_chance, float):
            raise TypeError("'repetition_chance' must be 'float'")
        if repetition_chance < 0.0 or repetition_chance > 1.0:
            raise ValueError("'repetition_chance' must be between 0.0 and 1.0")
        self._repetition_chance = repetition_chance

    ### PRIVATE PROPERTIES ###

    @property
    def _done(self) -> bool:
        r""":obj:`bool` indicating whether the process is done, which is when
        the mask is filled with ``None``'s.
        """
        if self._include_empty_measures:
            return all(item is None for item in self._mask)
        else:
            return all(item in (None, self._min_dynamic_number) for item in self._mask)

    @property
    def _min_dynamic_number(self) -> int:
        r"""Numerical representation of :attr:`min_dynamic`."""
        return self._dynamic_name_to_dynamic_ordinal[self._min_dynamic]



from typing import Any, Union

import abjad


class LeafDynMaker(abjad.LeafMaker):
    r"""Creates leaves and logical ties from :obj:`list`'s of pitches,
    durations, dynamics, and articulations. It is an extension of
    |abjad.LeafMaker| which can take optional :obj:`list`'s of dynamics and
    articulations.

    Basic usage:
        Usage is similar to |abjad.LeafMaker|:

        >>> pitches = [0, 2, 4, 5, 7, 9]
        >>> durations = [(1, 32), (2, 32), (3, 32), (4, 32), (5, 32), (6, 32)]
        >>> dynamics = ['pp', 'p', 'mp', 'mf', 'f', 'ff']
        >>> articulations = ['.', '>', '-', '_', '^', '+']
        >>> maker = auxjad.LeafDynMaker()
        >>> notes = maker(pitches, durations, dynamics, articulations)
        >>> staff = abjad.Staff(notes)
        >>> abjad.show(staff)

        ..  docs::

            \new Staff
            {
                c'32
                \pp
                -\staccato
                d'16
                \p
                -\accent
                e'16.
                \mp
                -\tenuto
                f'8
                \mf
                -\portato
                g'8
                \f
                -\marcato
                ~
                g'32
                a'8.
                \ff
                -\stopped
            }

        ..  figure:: ../_images/LeafDynMaker-iv96zmptned.png

    ``pitches``:
        Tuple elements in ``pitches`` result in chords. None-valued elements
        in ``pitches`` result in rests:

        >>> pitches = [5, None, (0, 2, 7)]
        >>> durations = [(1, 4), (1, 8), (1, 16)]
        >>> dynamics = ['p', None, 'f']
        >>> articulations = ['staccato', None, 'tenuto']
        >>> maker = auxjad.LeafDynMaker()
        >>> notes = maker(pitches, durations, dynamics, articulations)
        >>> staff = abjad.Staff(notes)
        >>> abjad.show(staff)

        ..  docs::

            \new Staff
            {
                f'4
                \p
                -\staccato
                r8
                <c' d' g'>16
                \f
                -\tenuto
            }

        ..  figure:: ../_images/LeafDynMaker-odpaavswk29.png

    ``omit_repeated_dynamics``:
        Can omit repeated dynamics with the keyword argument
        ``omit_repeated_dynamics``:

        >>> pitches = [0, 2, 4, 5, 7, 9]
        >>> durations = [(1, 32), (2, 32), (3, 32), (4, 32), (5, 32), (6, 32)]
        >>> dynamics = ['pp', 'pp', 'mp', 'f', 'f', 'p']
        >>> maker = auxjad.LeafDynMaker()
        >>> notes = maker(pitches,
        ...               durations,
        ...               dynamics,
        ...               omit_repeated_dynamics=True,
        ...               )
        >>> staff = abjad.Staff(notes)
        >>> abjad.show(staff)

        ..  docs::

            \new Staff
            {
                c'32
                \pp
                d'16
                e'16.
                \mp
                f'8
                \f
                g'8
                ~
                g'32
                a'8.
                \p
            }

        ..  figure:: ../_images/LeafDynMaker-jryhls4d6ic.png

    ``dynamics`` and ``articulations``:
        The lengths ``dynamics`` and ``articulations`` can be shorter than the
        lengths of ``pitches`` and ``durations`` (whatever is the greatest):

        >>> pitches = [0, 2, 4, 5, 7, 9, 11, 12]
        >>> durations = (1, 4)
        >>> dynamics = ['p', 'f', 'ff']
        >>> articulations = ['.', '>']
        >>> maker = auxjad.LeafDynMaker()
        >>> notes = maker(pitches, durations, dynamics, articulations)
        >>> staff = abjad.Staff(notes)
        >>> abjad.show(staff)

        ..  docs::

            \new Staff
            {
                c'4
                \p
                -\staccato
                d'4
                \f
                -\accent
                e'4
                \ff
                f'4
                g'4
                a'4
                b'4
                c''4
            }

        ..  figure:: ../_images/LeafDynMaker-1losyp9vc1di.png

    ``cyclic_dynamics`` and ``cyclic_articulations``:
        If the lengths of either ``dynamics`` and ``articulations`` are shorter
        than the lengths of ``pitches`` and ``durations`` (whatever is the
        greatest), use the optional keyword arguments ``cyclic_dynamics`` and
        ``cyclic_articulations`` to apply those parameters cyclically:

        >>> pitches = [0, 2, 4, 5, 7, 9, 11, 12]
        >>> durations = (1, 4)
        >>> dynamics = ['p', 'f', 'ff']
        >>> articulations = ['.', '>']
        >>> maker = auxjad.LeafDynMaker()
        >>> notes = maker(pitches,
        ...               durations,
        ...               dynamics,
        ...               articulations,
        ...               cyclic_dynamics=True,
        ...               cyclic_articulations=True,
        ...               )
        >>> staff = abjad.Staff(notes)
        >>> abjad.show(staff)

        ..  docs::

            \new Staff
            {
                c'4
                \p
                - \staccato
                d'4
                \f
                - \accent
                e'4
                \ff
                - \staccato
                f'4
                \p
                - \accent
                g'4
                \f
                - \staccato
                a'4
                \ff
                - \accent
                b'4
                \p
                - \staccato
                c''4
                \f
                - \accent
            }

        ..  figure:: ../_images/LeafDynMaker-x8zhyel42i.png

    ``articulations`` and ``dynamics`` of length ``1``:
        If the length of ``articulations`` or ``dynamics`` is ``1``, they will
        be applied only to the first element.

        >>> pitches = [0, 2, 4, 5, 7, 9, 11, 12]
        >>> durations = (1, 4)
        >>> dynamics = 'p'
        >>> articulations = '.'
        >>> maker = auxjad.LeafDynMaker()
        >>> notes = maker(pitches, durations, dynamics, articulations)
        >>> staff = abjad.Staff(notes)
        >>> abjad.show(staff)

        ..  docs::

            \new Staff
            {
                c'4
                \p
                -\staccato
                d'4
                e'4
                f'4
                g'4
                a'4
                b'4
                c''4
            }

        ..  figure:: ../_images/LeafDynMaker-wwl4pcpmwhv232n7j7ws.png

    ``cyclic_dynamics`` and ``cyclic_articulations`` of length ``1``:
        To apply them to all elements, use the ``cyclic_dynamics`` and
        ``cyclic_articulations`` optioanl keywords.

        >>> pitches = [0, 2, 4, 5, 7, 9, 11, 12]
        >>> durations = (1, 4)
        >>> dynamics = 'p'
        >>> articulations = '.'
        >>> maker = auxjad.LeafDynMaker()
        >>> notes = maker(pitches,
        ...               durations,
        ...               dynamics,
        ...               articulations,
        ...               cyclic_articulations=True,
        ...               )
        >>> staff = abjad.Staff(notes)
        >>> abjad.show(staff)

        ..  docs::

            \new Staff
            {
                c'4
                \p
                -\staccato
                d'4
                -\staccato
                e'4
                -\staccato
                f'4
                -\staccato
                g'4
                -\staccato
                a'4
                -\staccato
                b'4
                - \staccato
                c''4
                - \staccato
            }

        ..  figure:: ../_images/LeafDynMaker-s5ic146v4qb.png

    Accepted types:
        Similarly to Abjad's native classes, it accepts many types of elements
        in its input :obj:`list`'s:

        >>> pitches = [0,
        ...            "d'",
        ...            'E4',
        ...            abjad.NumberedPitch(5),
        ...            abjad.NamedPitch("g'"),
        ...            abjad.NamedPitch('A4'),
        ...            ]
        >>> durations = [(1, 32),
        ...              '2/32',
        ...              abjad.Duration('3/32'),
        ...              abjad.Duration(0.125),
        ...              abjad.Duration(5, 32),
        ...              abjad.Duration(6 / 32),
        ...              ]
        >>> dynamics = ['p',
        ...             abjad.Dynamic('f'),
        ...             ]
        >>> articulations = ['>',
        ...                  abjad.Articulation('-'),
        ...                  abjad.Articulation('.'),
        ...                  ]
        >>> maker = auxjad.LeafDynMaker()
        >>> notes = maker(pitches, durations, dynamics, articulations)
        >>> staff = abjad.Staff(notes)
        >>> abjad.show(staff)

        ..  docs::

            \new Staff
            {
                c'32
                \p
                - \accent
                d'16
                \f
                - \tenuto
                e'16.
                - \staccato
                f'8
                g'8
                ~
                g'32
                a'8.
            }

        ..  figure:: ../_images/LeafDynMaker-79n2uz1n8j3.png
    """

    ### CLASS VARIABLES ###

    __slots__ = (
        "_omit_repeated_dynamics",
        "_cyclic_dynamics",
        "_cyclic_articulations",
    )

    ### INITIALISER ###

    def __init__(self) -> None:
        super().__init__()

    ### SPECIAL METHODS ###

    def __call__(
        self,
        pitches,
        durations,
        dynamics=None,
        articulations=None,
        *,
        omit_repeated_dynamics: bool = False,
        cyclic_dynamics: bool = False,
        cyclic_articulations: bool = False,
    ) -> abjad.Selection:
        r"""Calls the leaf-maker on ``pitches``, ``durations``, ``dynamics``,
        and ``articulations``, returning an |abjad.Selection|.
        """
        if dynamics is not None:
            for dynamic in dynamics:
                if dynamic is not None:
                    if not isinstance(dynamic, (str, abjad.Dynamic)):
                        raise TypeError("dynamics must be 'str' or 'abjad.Dynamic'")
        if articulations is not None:
            for articulation in articulations:
                if articulation is not None:
                    if not isinstance(articulation, (str, abjad.Articulation)):
                        raise TypeError("articulations must be 'str' or 'abjad.Articulation'")
        if not isinstance(omit_repeated_dynamics, bool):
            raise TypeError("'omit_repeated_dynamics' must be 'bool")
        if not isinstance(cyclic_dynamics, bool):
            raise TypeError("'cyclic_dynamics' must be 'bool")
        if not isinstance(cyclic_articulations, bool):
            raise TypeError("'cyclic_articulations' must be 'bool")

        leaves = super().__call__(pitches, durations)
        dummy_container = abjad.Container(leaves)
        logical_ties = leaves.logical_ties()

        pitches_ = self._listify(pitches)
        durations_ = self._listify(durations)
        dynamics_ = self._listify(dynamics)
        articulations_ = self._listify(articulations)

        greatest_len = max(len(pitches_), len(durations_))
        self._fill_list(
            dynamics_,
            greatest_len,
            cyclic=cyclic_dynamics,
        )
        self._fill_list(
            articulations_,
            greatest_len,
            cyclic=cyclic_articulations,
        )

        self._add_dynamics_and_articulations(
            logical_ties,
            dynamics_,
            articulations_,
            omit_repeated_dynamics,
        )

        output = dummy_container[:]
        dummy_container[:] = []
        return output

    ### PRIVATE METHODS ###

    def _add_dynamics_and_articulations(
        self,
        logical_ties: list[abjad.LogicalTie],
        dynamics_: list[
            Union[
                str,
                abjad.Dynamic,
            ]
        ],
        articulations_: list[
            Union[
                str,
                abjad.Articulation,
            ]
        ],
        omit_repeated_dynamics: bool,
    ) -> None:
        r"""Adds dynamics and articulations to logical ties."""
        previous_dynamic = None
        for logical_tie, dynamic, articulation in zip(
            logical_ties,
            dynamics_,
            articulations_,
        ):
            if dynamic is not None and (not omit_repeated_dynamics or dynamic != previous_dynamic):
                abjad.attach(abjad.Dynamic(dynamic), logical_tie.head)
                previous_dynamic = dynamic
            if articulation is not None:
                if isinstance(articulation, str):
                    abjad.attach(
                        abjad.Articulation(articulation),
                        logical_tie.head,
                    )
                else:
                    abjad.attach(articulation, logical_tie.head)

    @staticmethod
    def _listify(
        argument: Any,
    ) -> list[Any]:
        r"""Returns a :obj:`list` if argument is not a :obj:`list`."""
        if argument:
            if isinstance(argument, list):
                return argument
            else:
                return [argument]
        else:
            return []

    @staticmethod
    def _fill_list(
        input_list: list[Any],
        length: int,
        *,
        cyclic: bool = False,
        default: Any = None,
    ) -> None:
        r"""Extends a :obj:`list` to a certain length, filling it with a
        default value. If ``cyclic`` is ``True``, then it fills the :obj:`list`
        by cycling the original ``input_list``.
        """
        if not cyclic:
            while len(input_list) < length:
                input_list.append(default)
        else:
            original_length = len(input_list)
            while len(input_list) < length:
                current_length = len(input_list)
                input_list.append(input_list[current_length % original_length])



import abjad

from .. import select
from .auto_rewrite_meter import auto_rewrite_meter


def sustain_notes(
    container: abjad.Container,
    *,
    sustain_multimeasure_rests: bool = True,
    rewrite_meter: bool = True,
) -> None:
    r"""Mutates an input container (of type |abjad.Container| or child class)
    in place and has no return value; this function will sustain all pitched
    leaves until the next pitched leaf, thus replacing all rests in between
    them.

    Basic usage:
        Simply call the function on a container.

        >>> staff = abjad.Staff(r"c'16 r8. d'16 r8. e'16 r8. f'16 r8.")
        >>> abjad.show(staff)

        ..  docs::

            \new Staff
            {
                c'16
                r8.
                d'16
                r8.
                e'16
                r8.
                f'16
                r8.
            }

        ..  figure:: ../_images/sustain_notes-w1e1pmruyce.png

        >>> auxjad.mutate.sustain_notes(staff)
        >>> abjad.show(staff)

        ..  docs::

            \new Staff
            {
                c'4
                d'4
                e'4
                f'4
            }

        ..  figure:: ../_images/sustain_notes-ythfpvkrvue.png

    ..  note::

        Auxjad automatically adds this function as an extension function to
        |abjad.mutate|. It can thus be used from either |auxjad.mutate|_ or
        |abjad.mutate| namespaces. Therefore, the two lines below are
        equivalent:

        >>> auxjad.mutate.close_containers(staff)
        >>> abjad.mutate.close_containers(staff)

    Leaves with same pitch:
        Leaves are sustained until the next pitched leaf, even if the pitch is
        the same.

        >>> staff = abjad.Staff(r"c'16 r8. c'16 r8. c'16 r8. c'16 r8.")
        >>> abjad.show(staff)

        ..  docs::

            \new Staff
            {
                c'16
                r8.
                c'16
                r8.
                c'16
                r8.
                c'16
                r8.
            }

        ..  figure:: ../_images/sustain_notes-oliqicqqw7q.png

        >>> auxjad.mutate.sustain_notes(staff)
        >>> abjad.show(staff)

        ..  docs::

            \new Staff
            {
                c'4
                c'4
                c'4
                c'4
            }

        ..  figure:: ../_images/sustain_notes-3alcbmhc2jt.png

    Consecutive leaves with the same pitch:
        Consecutive pitched leaves with a same pitch will not be tied.

        >>> staff = abjad.Staff(
        ...     r"<c' e'>16 r8. <c' e'>4 <c' e'>4 <c' e'>16 r8."
        ... )
        >>> abjad.show(staff)

        ..  docs::

            \new Staff
            {
                <c' e'>16
                r8.
                <c' e'>4
                <c' e'>4
                <c' e'>16
                r8.
            }

        ..  figure:: ../_images/sustain_notes-ek4ujjintt8.png

        >>> auxjad.mutate.sustain_notes(staff)
        >>> abjad.show(staff)

        ..  docs::

            \new Staff
            {
                <c' e'>4
                <c' e'>4
                <c' e'>4
                <c' e'>4
            }

        ..  figure:: ../_images/sustain_notes-f7au6hojq99.png

    Tuplets:
        This function handles tuplets.

        >>> staff = abjad.Staff(
        ...     r"\times 2/3 {c'4 d'4 r4} r8 e'8 \times 2/3 {f'8 r4}"
        ... )
        >>> abjad.show(staff)

        ..  docs::

            \new Staff
            {
                \times 2/3
                {
                    c'4
                    d'4
                    r4
                }
                r8
                e'8
                \times 2/3
                {
                    f'8
                    r4
                }
            }

        ..  figure:: ../_images/sustain_notes-nsjvhnyrkea.png

        >>> auxjad.mutate.sustain_notes(staff)
        >>> abjad.show(staff)

        ..  docs::

            \new Staff
            {
                \times 2/3
                {
                    c'4
                    d'2
                    ~
                }
                d'8
                e'8
                f'4
            }

        ..  figure:: ../_images/sustain_notes-26l9hob8wko.png

    Complex example:
        This function can handle containers with a mixture of notes, chords,
        and rests, as well as tuplets.

        >>> staff = abjad.Staff(
        ...     r"c'16 r8. d'16 r8. r8 r32 <e' g'>32 r16 r4 "
        ...     r"\times 2/3 {r4 f'4 r4} r4 g'8 r8 a'4 ~ "
        ...     r"a'16 r8. b'4 c''8 r8 "
        ...     r"r4. d''8 \times 4/5 {r8 d''2}"
        ... )
        >>> abjad.show(staff)

        ..  docs::

            \new Staff
            {
                c'16
                r8.
                d'16
                r8.
                r8
                r32
                <e' g'>32
                r16
                r4
                \times 2/3
                {
                    r4
                    f'4
                    r4
                }
                r4
                g'8
                r8
                a'4
                ~
                a'16
                r8.
                b'4
                c''8
                r8
                r4.
                d''8
                \times 4/5
                {
                    r8
                    d''2
                }
            }

        ..  figure:: ../_images/sustain_notes-cpw7dvpegge.png

        >>> auxjad.mutate.sustain_notes(staff)
        >>> abjad.show(staff)

        ..  docs::

            \new Staff
            {
                c'4
                d'4
                ~
                d'8
                ~
                d'32
                <e' g'>16.
                ~
                <e' g'>4
                ~
                \times 2/3
                {
                    <e' g'>4
                    f'2
                    ~
                }
                f'4
                g'4
                a'2
                b'4
                c''4
                ~
                c''4.
                d''8
                ~
                \times 4/5
                {
                    d''8
                    d''2
                }
            }

        ..  figure:: ../_images/sustain_notes-z8t2jwxsvar.png

    ``sustain_multimeasure_rests``:
        By default, notes are tied across multi-measure rests.

        >>> staff = abjad.Staff(
        ...     r"r4 c'16 r8. d'16 r4.. "
        ...     r"R1 "
        ...     r"r4 e'4 r2 "
        ...     r"\time 5/8 r8 f'4 r4 "
        ...     r"R1 * 5/8 "
        ...     r"r8 g'8 a'8 r4"
        ... )
        >>> abjad.show(staff)

        ..  docs::

            \new Staff
            {
                r4
                c'16
                r8.
                d'16
                r4..
                R1
                r4
                e'4
                r2
                \time 5/8
                r8
                f'4
                r4
                R1 * 5/8
                r8
                g'8
                a'8
                r4
            }

        ..  figure:: ../_images/sustain_notes-mJOOARIUAp.png

        >>> auxjad.mutate.sustain_notes(staff)
        >>> abjad.show(staff)

        ..  docs::

            \new Staff
            {
                r4
                c'4
                d'2
                ~
                d'1
                ~
                d'4
                e'2.
                ~
                \time 5/8
                e'8
                f'2
                ~
                f'4.
                ~
                f'4
                ~
                f'8
                g'8
                a'4.
            }

        ..  figure:: ../_images/sustain_notes-iLTiWERSvO.png

        To disable sustaining across those, set ``sustain_multimeasure_rests``
        to  ``False``:

        >>> auxjad.mutate.sustain_notes(staff, sustain_multimeasure_rests=True)
        >>> abjad.show(staff)

        ..  docs::

            \new Staff
            {
                r4
                c'4
                d'2
                R1
                r4
                e'2.
                ~
                \time 5/8
                e'8
                f'2
                R1 * 5/8
                r8
                g'8
                a'4.
            }

        ..  figure:: ../_images/sustain_notes-9WeilArLex.png

    ``rewrite_meter``:
        By default, |auxjad.mutate.auto_rewrite_meter()| is summoned after
        notes are sustained.

        >>> staff = abjad.Staff(
        ...     r"r4 c'16 r8. d'16 r4.. "
        ...     r"R1"
        ...     r"r4 e'4 r2"
        ...     r"\time 5/8 r8 f'4 r4"
        ...     r"R1 * 5/8 "
        ...     r"r8 g'8 a'8 r4"
        ... )
        >>> abjad.show(staff)

        ..  docs::

            \new Staff
            {
                r4
                c'16
                r8.
                d'16
                r4..
                R1
                r4
                e'4
                r2
                \time 5/8
                r8
                f'4
                r4
                R1 * 5/8
                r8
                g'8
                a'8
                r4
            }

        ..  figure:: ../_images/sustain_notes-Egibb7mRfJ.png

        To disable this behaviour, set ``rewrite_meter`` to ``False``:

        >>> auxjad.mutate.sustain_notes(staff, rewrite_meter=False)
        >>> abjad.show(staff)

        ..  docs::

            \new Staff
            {
                r4
                c'16
                ~
                c'8.
                d'16
                ~
                d'4..
                ~
                d'1
                ~
                d'4
                e'4
                ~
                e'2
                ~
                \time 5/8
                e'8
                f'4
                ~
                f'4
                ~
                f'2
                ~
                f'8
                ~
                f'8
                g'8
                a'8
                ~
                a'4
            }

        ..  figure:: ../_images/sustain_notes-iTuoVMoWgm.png

    ..  warning::

        The input container must be a contiguous logical voice. When dealing
        with a container with multiple subcontainers (e.g. a score containing
        multiple staves), the best approach is to cycle through these
        subcontainers, applying this function to them individually.
    """
    if not isinstance(container, abjad.Container):
        raise TypeError("argument must be 'abjad.Container' or child class")
    if not abjad.select(container).leaves().are_contiguous_logical_voice():
        raise ValueError("argument must be contiguous logical voice")
    if not isinstance(sustain_multimeasure_rests, bool):
        raise TypeError("'sustain_multimeasure_rests' must be 'bool'")
    if not isinstance(rewrite_meter, bool):
        raise TypeError("'rewrite_meter' must be 'bool'")

    logical_selections = select.logical_selections(
        container,
        include_multimeasure_rests=sustain_multimeasure_rests,
    )
    pitch = None
    last_leaf = None
    for logical_selection in logical_selections:
        first_leaf = abjad.select(logical_selection).leaf(0)
        # if pitched
        if isinstance(first_leaf, abjad.Note):
            pitch = first_leaf.written_pitch
            last_leaf = abjad.select(logical_selection).leaf(-1)
            continue
        elif isinstance(first_leaf, abjad.Chord):
            # needs 2D list as abjad.LeafMaker considers a single pitch list as
            # a series of consecutive pitches, but any sublist as a chord
            pitch = [[pitch for pitch in first_leaf.written_pitches]]
            last_leaf = abjad.select(logical_selection).leaf(-1)
            continue
        # if not pitched
        for leaf in abjad.select(logical_selection).leaves():
            if last_leaf is None:
                break
            elif isinstance(leaf, abjad.Rest):
                duration = leaf.written_duration
            elif isinstance(leaf, abjad.MultimeasureRest):
                if not sustain_multimeasure_rests:
                    pitch = None
                    last_leaf = None
                    break
                duration = abjad.get.duration(leaf)
            else:
                raise TypeError(
                    "Leaves in 'container' must be notes, chords, rests, or multi-measure rests"
                )
            if abjad.get.indicator(last_leaf, abjad.Tie) is None:
                abjad.attach(abjad.Tie(), last_leaf)
            new_leaves = abjad.LeafMaker()(pitch, duration)
            for indicator in abjad.get.indicators(leaf):
                abjad.attach(
                    indicator,
                    abjad.select(new_leaves).leaf(0),
                )
            abjad.mutate.replace(leaf, new_leaves)
            last_leaf = abjad.select(new_leaves).leaf(-1)

    # rewriting meter
    if rewrite_meter:
        auto_rewrite_meter(container)



from textwrap import dedent
from typing import Union

import abjad
from abjad import piano_pedal as piano_pedal_


def piano_pedal(
    argument: Union[abjad.Component, abjad.Selection],
    *,
    half_pedal: bool = False,
    until_the_end: bool = False,
    omit_raise_pedal_glyph: bool = False,
    selector: abjad.Expression = abjad.select().leaves(),
    start_piano_pedal: abjad.StartPianoPedal = None,
    stop_piano_pedal: abjad.StopPianoPedal = None,
    tag: abjad.Tag = None,
) -> None:
    r"""Attaches piano pedal indicators. This function extends the capabilities
    of Abjad's built-in |abjad.piano_pedal()|.

    Basic usage:
        Basic usage is identical to Abjad's built-ind |abjad.piano_pedal()|:

        >>> staff = abjad.Staff(r"c'4 d'4 e'4 f'4")
        >>> abjad.piano_pedal(staff[:])
        >>> abjad.show(staff)

        ..  docs::

            \new Staff
            {
                c'4
                \sustainOn
                d'4
                e'4
                f'4
                \sustainOff
            }

        ..  figure:: ../_images/piano_pedal-KaleXflNvL.png

    ..  note::

        Auxjad automatically replaces Abjad's built-in |abjad.piano_pedal()|
        with this function. Therefore it can be used either as
        :func:`abjad.piano_pedal()` or |abjad.piano_pedal()|, as shown below:

        >>> staff1 = abjad.Staff(r"c'4 d'4 e'4 f'4")
        >>> staff2 = abjad.Staff(r"c'4 d'4 e'4 f'4")
        >>> abjad.piano_pedal(staff1[:])
        >>> auxjad.piano_pedal(staff2[:])
        >>> selections = [staff1[:], staff2[:]]
        >>> auxjad.get.selections_are_identical(selections)
        True

    ``until_the_end``:
        Call the function with ``until_the_end`` set to ``True`` to add an
        arrow to the initial pedal glyph:

        >>> staff = abjad.Staff(r"c'4 d'4 e'4 f'4")
        >>> abjad.piano_pedal(staff[:],
        ...                   until_the_end=True,
        ...                   )
        >>> abjad.show(staff)

        ..  docs::

            \new Staff
            {
                \once \override Staff.SustainPedal.stencil =
                    #(lambda (grob) (grob-interpret-markup grob
                        #{
                            \markup {
                                \concat {
                                    \musicglyph "pedal.Ped"
                                    \musicglyph "pedal.."
                                }
                                \raise #-0.3 "→"
                            }
                        #}))
                c'4
                \sustainOn
                d'4
                e'4
                f'4
                \sustainOff
            }

        ..  figure:: ../_images/piano_pedal-rA7ZHeMrjf.png

    ``omit_raise_pedal_glyph``:
        Call the function with ``omit_raise_pedal_glyph`` set to ``True`` to
        remove the raise pedal glyph:

        >>> staff = abjad.Staff(r"c'4 d'4 e'4 f'4")
        >>> abjad.piano_pedal(staff[:],
        ...                   omit_raise_pedal_glyph=True,
        ...                   )
        >>> abjad.show(staff)

        ..  docs::

            \new Staff
            {
                c'4
                \sustainOn
                d'4
                e'4
                \once \override Staff.SustainPedal.stencil = ##f
                f'4
                \sustainOff
            }

        ..  figure:: ../_images/piano_pedal-p8S1KwHLIx.png

        Combined with ``until_the_end=True``:

        >>> staff = abjad.Staff(r"c'4 d'4 e'4 f'4")
        >>> abjad.piano_pedal(staff[:],
        ...                   until_the_end=True,
        ...                   omit_raise_pedal_glyph=True,
        ...                   )
        >>> abjad.show(staff)

        ..  docs::

            \new Staff
            {
                \once \override Staff.SustainPedal.stencil =
                    #(lambda (grob) (grob-interpret-markup grob
                        #{
                            \markup {
                                \concat {
                                    \musicglyph "pedal.Ped"
                                    \musicglyph "pedal.."
                                }
                                \raise #-0.3 "→"
                            }
                        #}))
                c'4
                \sustainOn
                d'4
                e'4
                \once \override Staff.SustainPedal.stencil = ##f
                f'4
                \sustainOff
            }

        ..  figure:: ../_images/piano_pedal-Fb5rE6QB1f.png

    ``half_pedal``:

        Call the function with ``half_pedal`` set to ``True`` to use the half
        pedalling glyph:

        >>> staff = abjad.Staff(r"c'4 d'4 e'4 f'4")
        >>> abjad.piano_pedal(staff[:],
        ...                   half_pedal=True,
        ...                   )
        >>> abjad.show(staff)

        ..  docs::

            \new Staff
            {
                \once \override Staff.SustainPedal.stencil =
                    #(lambda (grob) (grob-interpret-markup grob
                        #{
                            \markup {
                                \larger "½"
                                \concat {
                                    \musicglyph "pedal.Ped"
                                    \musicglyph "pedal.."
                                }
                            }
                        #}))
                c'4
                \sustainOn
                d'4
                e'4
                f'4
                \sustainOff
            }

        ..  figure:: ../_images/piano_pedal-EYs495mi9K.png

        Combined with ``until_the_end=True``:

        >>> staff = abjad.Staff(r"c'4 d'4 e'4 f'4")
        >>> abjad.piano_pedal(staff[:],
        ...                   half_pedal=True,
        ...                   until_the_end=True,
        ...                   )
        >>> abjad.show(staff)

        ..  docs::

            \new Staff
            {
                \once \override Staff.SustainPedal.stencil =
                    #(lambda (grob) (grob-interpret-markup grob
                        #{
                            \markup {
                                \larger "½"
                                \concat {
                                    \musicglyph "pedal.Ped"
                                    \musicglyph "pedal.."
                                }
                                \raise #-0.3 "→"
                            }
                        #}))
                c'4
                \sustainOn
                d'4
                e'4
                f'4
                \sustainOff
            }

        ..  figure:: ../_images/piano_pedal-viOZdVeRs9.png

    Pedal style:
        The style of the sustain pedal can be tweaked using |abjad.setting()|
        as shown below:

        >>> staff = abjad.Staff(r"c'4 d'4 e'4 f'4")
        >>> abjad.piano_pedal(staff[:])
        >>> abjad.setting(staff).pedal_sustain_style = "#'mixed"
        >>> abjad.show(staff)

        ..  docs::

            \new Staff
            \with
            {
                pedalSustainStyle = #'mixed
            }
            {
                c'4
                \sustainOn
                d'4
                e'4
                f'4
                \sustainOff
            }

        ..  figure:: ../_images/piano_pedal-6q9Swb2elq.png

        This tweak also works with ``until_the_end=True`` if desired:

        >>> staff = abjad.Staff(r"c'4 d'4 e'4 f'4")
        >>> abjad.piano_pedal(staff[:],
        ...                   until_the_end=True,
        ...                   )
        >>> abjad.setting(staff).pedal_sustain_style = "#'mixed"
        >>> abjad.show(staff)

        ..  docs::

            \new Staff
            \with
            {
                pedalSustainStyle = #'mixed
            }
            {
                \once \override Staff.SustainPedal.stencil =
                    #(lambda (grob) (grob-interpret-markup grob
                        #{
                            \markup {
                                \concat {
                                    \musicglyph "pedal.Ped"
                                    \musicglyph "pedal.."
                                }
                                \raise #-0.3 "→"
                            }
                        #}))
                c'4
                \sustainOn
                d'4
                e'4
                f'4
                \sustainOff
            }

        ..  figure:: ../_images/piano_pedal-cYyRwFvnFH.png
    """
    assert isinstance(selector, abjad.Expression)
    argument = selector(argument)
    leaves = abjad.Selection(argument).leaves()
    start_leaf = leaves[0]
    stop_leaf = leaves[-1]
    if not half_pedal and until_the_end:
        string = r"""
            \once \override Staff.SustainPedal.stencil =
                #(lambda (grob) (grob-interpret-markup grob
                    #{
                        \markup {
                            \concat {
                                \musicglyph "pedal.Ped"
                                \musicglyph "pedal.."
                            }
                            \raise #-0.3 "→"
                        }
                    #}))
            """
    elif half_pedal and not until_the_end:
        string = r"""
            \once \override Staff.SustainPedal.stencil =
                #(lambda (grob) (grob-interpret-markup grob
                    #{
                        \markup {
                            \larger "½"
                            \concat {
                                \musicglyph "pedal.Ped"
                                \musicglyph "pedal.."
                            }
                        }
                    #}))
            """
    elif half_pedal and until_the_end:
        string = r"""
            \once \override Staff.SustainPedal.stencil =
                #(lambda (grob) (grob-interpret-markup grob
                    #{
                        \markup {
                            \larger "½"
                            \concat {
                                \musicglyph "pedal.Ped"
                                \musicglyph "pedal.."
                            }
                            \raise #-0.3 "→"
                        }
                    #}))
            """
    if half_pedal or until_the_end:
        string = dedent(string).strip()
        pedal_tweak = abjad.LilyPondLiteral(string)
        abjad.attach(pedal_tweak, start_leaf)
    if omit_raise_pedal_glyph:
        omit_raise_pedal_glyph_tweak = abjad.LilyPondLiteral(
            r"\once \override Staff.SustainPedal.stencil = ##f"
        )
        abjad.attach(omit_raise_pedal_glyph_tweak, stop_leaf)
    piano_pedal_(
        argument=argument,
        selector=selector,
        start_piano_pedal=start_piano_pedal,
        stop_piano_pedal=stop_piano_pedal,
        tag=tag,
    )


### MONKEY PATCHING ###

abjad.spanners.piano_pedal = piano_pedal
abjad.piano_pedal = piano_pedal



"""
mutate
======

Auxjad's mutation functions. These can be accesses via either:

>>> abjad.mutate.auxjad_function(container)
>>> auxjad.mutate.auxjad_function(container)
"""

import abjad

from .auto_rewrite_meter import auto_rewrite_meter
from .close_container import close_container
from .contract_notes import contract_notes
from .enforce_time_signature import enforce_time_signature
from .extend_notes import extend_notes
from .extract_trivial_tuplets import extract_trivial_tuplets
from .fill_with_rests import fill_with_rests
from .merge_hairpins import merge_hairpins
from .merge_partial_tuplets import merge_partial_tuplets
from .prettify_rewrite_meter import prettify_rewrite_meter
from .remove_repeated_dynamics import remove_repeated_dynamics
from .remove_repeated_time_signatures import remove_repeated_time_signatures
from .reposition_clefs import reposition_clefs
from .reposition_dynamics import reposition_dynamics
from .reposition_slurs import reposition_slurs
from .respell_augmented_unisons import respell_augmented_unisons
from .rests_to_multimeasure_rest import rests_to_multimeasure_rest
from .sustain_notes import sustain_notes
from .sync_containers import sync_containers

### EXTENSION FUNCTIONS ###

abjad.mutate.auto_rewrite_meter = auto_rewrite_meter
abjad.mutate.close_container = close_container
abjad.mutate.contract_notes = contract_notes
abjad.mutate.enforce_time_signature = enforce_time_signature
abjad.mutate.extend_notes = extend_notes
abjad.mutate.extract_trivial_tuplets = extract_trivial_tuplets
abjad.mutate.fill_with_rests = fill_with_rests
abjad.mutate.merge_hairpins = merge_hairpins
abjad.mutate.merge_partial_tuplets = merge_partial_tuplets
abjad.mutate.prettify_rewrite_meter = prettify_rewrite_meter
abjad.mutate.remove_repeated_dynamics = remove_repeated_dynamics
abjad.mutate.remove_repeated_time_signatures = remove_repeated_time_signatures
abjad.mutate.reposition_clefs = reposition_clefs
abjad.mutate.reposition_dynamics = reposition_dynamics
abjad.mutate.reposition_slurs = reposition_slurs
abjad.mutate.respell_augmented_unisons = respell_augmented_unisons
abjad.mutate.rests_to_multimeasure_rest = rests_to_multimeasure_rest
abjad.mutate.sustain_notes = sustain_notes
abjad.mutate.sync_containers = sync_containers



from typing import Optional, Union

import abjad


def _generate_harmonics_from_pitch(
    fundamental: abjad.Pitch,
    upper_bound: abjad.Pitch,
) -> list:
    r"""Private function used by |auxjad.get.virtual_fundamental()|. Generates
    the harmonic series for a given fundamental pitch up to a given upper
    bound.
    """
    harmonic_intervals = [12, 7, 5, 4, 3, 3, 2, 2, 2, 1, 2, 1, 2, 1, 1]
    harmonics = [fundamental]
    for interval in harmonic_intervals:
        harmonic = harmonics[-1] + interval
        if harmonic > upper_bound:
            break
        harmonics.append(harmonic)
    else:  # if upper bound not reached yet, harmonics go in semitones
        while harmonics[-1] < upper_bound:
            harmonic = harmonics[-1] + 1
            harmonics.append(harmonic)
    return harmonics


def virtual_fundamental(
    pitches: Union[abjad.PitchSegment, abjad.Chord],
    *,
    min_fundamental: Optional[abjad.Pitch] = None,
) -> abjad.NamedPitch:
    r"""Returns the virtual fundamental (highest common fundamental) for all
    pitches in a |abjad.PitchSegment| or |abjad.Chord|. Return value is of type
    |abjad.NamedPitch|.

    Basic usage:
        This function will look for the highest fundamental whose harmonic
        series contains all notes in a given collection of pitches. E.g.:

        >>> pitches = abjad.PitchSegment(r"c'' g''")
        >>> auxjad.get.virtual_fundamental(pitches)
        c'

        >>> pitches = abjad.PitchSegment(r"c'' e'' g''")
        >>> auxjad.get.virtual_fundamental(pitches)
        c

        >>> pitches = abjad.PitchSegment(r"c'' f'' g''")
        >>> auxjad.get.virtual_fundamental(pitches)
        f,

        >>> pitches = abjad.PitchSegment(r"c'' d'' ef'' fs''")
        >>> auxjad.get.virtual_fundamental(pitches)
        bf,,

    ..  note::

        Auxjad automatically adds this function as an extension function to
        |abjad.get|. It can thus be used from either |auxjad.get|_ or
        |abjad.get| namespaces. Therefore, the two lines below are equivalent:

        >>> pitches = abjad.PitchSegment(r"c'' g''")
        >>> auxjad.get.virtual_fundamental(pitches)
        c'
        >>> abjad.get.virtual_fundamental(pitches)
        c'

    Input types:
        Input types can be |abjad.PitchSegment| or |abjad.Chord|:

        >>> pitches = abjad.PitchSegment(r"c'' cs'' d'' ef'' e'' fs''")
        >>> auxjad.get.virtual_fundamental(pitches)
        d,,
        >>> chord = abjad.Chord(r"<c'' cs'' d'' ef'' e'' fs''>4")
        >>> auxjad.get.virtual_fundamental(chord)
        d,,
        >>> staff = abjad.Staff(r"r4 <c'' cs'' d'' ef'' e'' fs''>4 r4")
        >>> auxjad.get.virtual_fundamental(staff[1])
        d,,

    ``min_fundamental``:
        The partials above the 13th partial can be approximated to include all
        chromatic pitches. This means that any complex of pitches will have a
        virtual fundamental, even though they are fairly distant from it. To
        limit the search of a fundamental by a minimum value, set
        ``min_fundamental`` to a specific pitch:

        >>> pitches = abjad.PitchSegment(r"c'' cs'' d'' ef'' e'' fs''")
        >>> auxjad.get.virtual_fundamental(
        ...     pitches,
        ...     min_fundamental=abjad.NamedPitch(r"c,,,"),
        ... )
        d,,

        >>> pitches = abjad.PitchSegment(r"c'' cs'' d'' ef'' e'' fs''")
        >>> auxjad.get.virtual_fundamental(
        ...     pitches,
        ...     min_fundamental=abjad.NumberedPitch(-48),
        ... )
        d,,

    ..  error::

        If a fundamental is not found with a given ``min_fundamental``, the
        function will raise a :exc:`ValueError` exception:

        >>> pitches = abjad.PitchSegment(r"c'' cs'' d'' ef'' e'' fs''")
        >>> auxjad.get.virtual_fundamental(
        ...     pitches,
        ...     min_fundamental=abjad.NamedPitch(r"c'"),
        ... )
        ValueError: No fundamental found above c'
    """
    if not isinstance(pitches, (abjad.PitchSegment, abjad.Chord)):
        raise TypeError("argument must be 'abjad.PitchSegment' or an 'abjad.Chord'")
    if min_fundamental is not None:
        if not isinstance(min_fundamental, (abjad.Pitch)):
            raise TypeError("'min_fundamental' must be 'abjad.Pitch' or child class")
    if isinstance(pitches, abjad.Chord):
        pitches = pitches.written_pitches
    fundamental = abjad.NumberedPitch(min(pitches))
    if min_fundamental is not None:
        min_fundamental = abjad.NumberedPitch(min_fundamental)
    while True:
        harmonics = _generate_harmonics_from_pitch(fundamental, max(pitches))
        if all(abjad.NumberedPitch(pitch) in harmonics for pitch in pitches):
            return abjad.NamedPitch(fundamental)
        fundamental = fundamental - 1
        if min_fundamental is not None and fundamental < min_fundamental:
            raise ValueError("No fundamental found above " + format(min_fundamental))
